# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

import numpy as np
from pyiron_atomistics.atomistics.master.parallel import ParallelMaster
from pyiron_base import JobGenerator

__author__ = "Jan Janssen"
__copyright__ = "Copyright 2020, Max-Planck-Institut für Eisenforschung GmbH - " \
                "Computational Materials Design (CM) Department"
__version__ = "1.0"
__maintainer__ = "Jan Janssen"
__email__ = "janssen@mpie.de"
__status__ = "development"
__date__ = "Sep 1, 2018"


def random_displacement(basis, displacement=0.05):
    basis_copy = basis.copy()
    random_vec = (np.random.random([len(basis_copy), 3]) - 0.5) * displacement
    basis_copy.positions += random_vec
    return basis_copy


class DisplacementJobGenerator(JobGenerator):
    @property
    def parameter_list(self):
        """

        Returns:
            (list)
        """
        return list(range(self._job.input['num_points']))

    @staticmethod
    def job_name(parameter):
        return "rsdis_job_" + str(parameter)

    def modify_job(self, job, parameter):
        job.structure = random_displacement(basis=job.structure, displacement=float(self._job.input['displacement']))
        return job


class RandomDisMaster(ParallelMaster):
    """

    Args:
        project:
        job_name:
    """

    def __init__(self, project, job_name):
        super(RandomDisMaster, self).__init__(project, job_name)
        self.__name__ = 'RandomDisMaster'
        self.__version__ = '0.0.1'
        self.input['num_points'] = (100, 'number of points')
        self.input['displacement'] = (0.05, 'displacement')
        self._job_generator = DisplacementJobGenerator(self)

    def collect_output(self):
        """

        Returns:

        """
        pass


class RandomMDJobGenerator(JobGenerator):
    @property
    def parameter_list(self):
        """

        Returns:
            (list)
        """
        return list(enumerate([int(job_id) for job_id in self._job.input['job_ids'].split()]))

    @staticmethod
    def job_name(parameter):
        return "md_job_" + str(parameter[1]) + '_' + str(parameter[0])

    def modify_job(self, job, parameter):
        job.structure = self._job.project.load(parameter[1]).structure
        job.server.accept_crash = True
        return job


class RandomMDMaster(ParallelMaster):
    """

    Args:
        project:
        job_name:
    """

    def __init__(self, project, job_name):
        super(RandomMDMaster, self).__init__(project, job_name)
        self.__name__ = 'RandomMDMaster'
        self.__version__ = '0.0.1'
        self.input['num_points'] = (100, 'number of points')
        self._job_generator = RandomMDJobGenerator(self)

    @property
    def structure_job_id_lst(self):
        return [int(job_id) for job_id in self.input['job_ids'].split()]

    @structure_job_id_lst.setter
    def structure_job_id_lst(self, job_id_lst):
        self.input['job_ids'] = ' '.join([str(job_id) for job_id in job_id_lst])

    def run_static(self):
        if 'job_ids' not in self.input._dataset['Parameter']:
            self.input['job_ids'] = ' '.join([str(job_id) for job_id in np.random.choice(self.structure_job_id_lst,
                                                                                         self.input['num_points'])])
            self.to_hdf()
        super(RandomMDMaster, self).run_static()

    def collect_output(self):
        """

        Returns:

        """
        pass



import os
import importlib
from pyiron_atomistics.lammps.interactive import LammpsInteractive
from pyiron_base import state

try:  # mpi4py is only supported on Linux and Mac Os X
    from pylammpsmpi import LammpsLibrary
except ImportError:
    pass


class LammpsInteractiveWithoutOutput(LammpsInteractive):
    def __init__(self, project, job_name):
        if not state.database.database_is_disabled:
            raise RuntimeError(
                "To run a `Without` job, the database must first be disabled. Please "
                "`from pyiron_base import state; "
                "state.update({'disable_database': True})`, and try again."
            )
        super().__init__(project, job_name)
        self._interactive_disable_log_file = True

    def to_hdf(self, hdf=None, group_name=None):
        """

        Args:
            hdf:
            group_name:

        Returns:

        """
        if not self._interactive_disable_log_file:
            super(LammpsInteractiveWithoutOutput, self).to_hdf(
                hdf=hdf, group_name=group_name
            )

    def interactive_flush(self, path="interactive", include_last_step=False):
        if not self._interactive_disable_log_file:
            super(LammpsInteractiveWithoutOutput, self).interactive_flush(
                path=path, include_last_step=include_last_step
            )

    def interactive_initialize_interface(self):
        if not self._interactive_disable_log_file:
            self._create_working_directory()
        if self.server.run_mode.interactive and self.server.cores == 1:
            lammps = getattr(importlib.import_module("lammps"), "lammps")
            if self._log_file is None:
                self._log_file = os.path.join(self.working_directory, "log.lammps")
            if not self._interactive_disable_log_file:
                self._interactive_library = lammps(
                    cmdargs=["-screen", "none", "-log", self._log_file],
                    comm=self._interactive_mpi_communicator,
                )
            else:
                self._interactive_library = lammps(
                    cmdargs=["-screen", "none", "-log", "none"],
                    comm=self._interactive_mpi_communicator,
                )
        else:
            self._interactive_library = LammpsLibrary(
                cores=self.server.cores, working_directory=self.working_directory
            )
        if not all(self.structure.pbc):
            self.input.control["boundary"] = " ".join(
                ["p" if coord else "f" for coord in self.structure.pbc]
            )
        self._reset_interactive_run_command()
        self.interactive_structure_setter(self.structure)

    def interactive_close(self):
        if not self._interactive_disable_log_file:
            super().interactive_close()

    def refresh_job_status(self):
        if not self._interactive_disable_log_file:
            super().refresh_job_status()



import posixpath, os, time

from scipy import optimize
from pyiron_base import InteractiveBase

from pyiron_contrib.atomistics.atomicrex.general_input import ScipyAlgorithm
from pyiron_contrib.atomistics.atomicrex.base import AtomicrexBase

import_success = False
try:
    import atomicrex

    import_success = True
except ImportError:
    pass


class AtomicrexInteractive(AtomicrexBase, InteractiveBase):
    def __init__(self, project, job_name):
        super().__init__(project, job_name)
        if import_success:
            self._interactive_library = atomicrex.Job()
        else:
            self._interactive_library = None
        self._read_input_files = False

    @property
    def atomicrex_job_object(self):
        return self._interactive_library

    def interactive_prepare_job(self):
        """
        Writes input files and calls necessary functions of the underlying atomicrex.Job class.
        """
        # Reading the input file again causes several issues
        if not self._read_input_files:
            if not os.path.isdir(self.path):
                os.makedirs(self.path)
            os.chdir(self.path)
            self.write_input(directory=self.path)
            input_file = "main.xml"
            self._interactive_library.parse_input_file(input_file)
            self._read_input_files = True
        self._interactive_library.prepare_fitting()
        # self._interactive_library.set_verbosity(2)

    def interactive_add_structure(identifier, structure, forces=None, params=None):
        """
        This should be done when the FlattenedARProperty is reworked to use the new FlattenedStorage,
        which allows to resize the necessary arrays on the fly.

        Wrapper around the atomicrex.Job add_ase_structure and add_library_structure methods

        Args:
            identifier ([type]): [description]
            structure ([type]): [description]
            params ([type], optional): [description]. Defaults to None.
        """
        raise NotImplementedError(
            "Changes needed in the atomicrex class before this can be implemented"
        )

    def interactive_calculate_residual(self):
        """
        Calculate the residual. prepare_job needs to be called first
        """
        return self._interactive_library.calculate_residual()

    def interactive_calculate_hessian(self, parameters=None, eps=0.0001):
        """
        Calculate the hessian. prepare_job needs to be called first
        """
        return self._interactive_library.calculate_hessian(
            parameters=parameters, eps=eps
        )

    def interactive_calculate_gradient(self, parameters=None, eps=0.0001):
        """
        Calculate the gradient. prepare_job needs to be called first
        """
        return self._interactive_library.calculate_gradient(
            parameters=parameters, eps=eps
        )

    def run_if_interactive(self):
        self.interactive_prepare_job()
        if isinstance(self.input.fit_algorithm, ScipyAlgorithm):
            self._scipy_run()
            # sleep between running and collecting so atomicrex output is flushed to file
            ## close to flush outputs to file
            self.interactive_close()
            self._scipy_collect(cwd=self.path)
        else:
            self._interactive_library.perform_fitting()
            ## close to flush outputs to file
            self.interactive_close()
            self.collect_output(cwd=self.path)

    def _scipy_run(self):
        if self.input.fit_algorithm.global_minimizer is None:
            res = optimize.minimize(
                fun=self._interactive_library.calculate_residual,
                x0=self._interactive_library.get_potential_parameters(),
                **self.input.fit_algorithm.local_minimizer_kwargs,
            )
        else:
            minimizer_func = optimize.__getattribute__(
                self.input.fit_algorithm.global_minimizer
            )
            res = minimizer_func(
                func=self._interactive_library.calculate_residual,
                **self.input.fit_algorithm.global_minimizer_kwargs,
            )

        #self._interactive_library.set_potential_parameters(res.x)
        self.output.residual = self._interactive_library.calculate_residual()
        self.output.iterations = res.nit
        self._interactive_library.print_potential_parameters()
        self._interactive_library.print_properties()
        self._interactive_library.output_results()
        return res

    def _scipy_collect(self, cwd=None):
        """
        Internal function that parses the output of an atomicrex job
        fitted using scipy.
        """
        if cwd is None:
            cwd = self.working_directory
        if self.input.__version__ >= "0.1.0":
            filepath = f"{cwd}/atomicrex.out"

        params_triggered = False
        structures_triggered = False

        with open(filepath, "r") as f:
            final_parameter_lines = []
            final_property_lines = []

            for l in f:
                if l.startswith("ERROR"):
                    self.status.aborted = True
                    self.output.error = l

                else:
                    if params_triggered:
                        if not l.startswith("---"):
                            final_parameter_lines.append(l)
                        else:
                            # Collecting lines with final parameters finished, hand over to the potential class
                            self.potential._parse_final_parameters(
                                final_parameter_lines
                            )
                            params_triggered = False

                    elif l.startswith("Potential parameters"):
                        # Get the number of dofs
                        n_fit_dofs = int(l.split("=")[1][:-3])
                        params_triggered = True

                    elif structures_triggered:
                        if not l.startswith("---"):
                            final_property_lines.append(l)
                        else:
                            # Collecting structure information finished, hand over structures class
                            self.structures._parse_final_properties(
                                final_property_lines
                            )
                            structures_triggered = False

                    elif l.startswith("Computing"):
                        structures_triggered = True
        self.status.finished = True
        self.to_hdf()



from __future__ import print_function
# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

from pyiron_base import GenericJob
import numpy as np
import matplotlib.pyplot as plt
from os.path import isfile
from glob import iglob
from os.path import abspath

from pyiron_contrib.image.image import Image
from pyiron_contrib.image.utils import DistributingList
from pyiron_base import DataContainer

from pyiron_contrib.image.custom_filters import brightness_filter

"""
Store and process image data within the pyiron framework. Functionality of the `skimage` library is automatically 
scraped, along with some convenience decorators to switch their function-based library to a class-method-based library.
"""

__author__ = "Liam Huber"
__copyright__ = "Copyright 2019, Max-Planck-Institut für Eisenforschung GmbH " \
                "- Computational Materials Design (CM) Department"
__version__ = "0.0"
__maintainer__ = "Liam Huber"
__email__ = "huber@mpie.de"
__status__ = "development"
__date__ = "Jan 30, 2020"


class ImageJob(GenericJob):
    """
    A job type for storing and processing image data.

    TODO: Consider allowing the `data` field of each image to be saved to hdf5...

    Attributes:
        images (DistributingList): A list of `Image` objects.
    """

    def __init__(self, project, job_name):
        super(ImageJob, self).__init__(project, job_name)
        self.__name__ = "ImageJob"
        self._images = DistributingList()
        self.input = DataContainer(table_name ="input")
        self.output = DataContainer(table_name ="output")


    @property
    def images(self):
        return self._images

    @images.setter
    def images(self, val):
        if isinstance(val, DistributingList):
            self._images = val
        elif isinstance(val, (tuple, list, np.ndarray)):
            if not all([isinstance(obj, Image) for obj in val]):
                raise ValueError("Only `Image`-type objects can be set to the `images` attribute.")
            self._images = DistributingList(val)
        else:
            raise ValueError("Images was expecting a list-like object, but got {}".format(type(val)))

    @staticmethod
    def _get_factors(n):
        i = int(n**0.5 + 0.5)
        while n % i != 0:
            i -= 1
        return i, int(n/i)

    def plot(self, mask=None, subplots_kwargs=None, imshow_kwargs=None, hide_axes=True):
        """
        Make a simple matplotlib `imshow` plot for each of the images on a grid.

        Args:
            mask (list/numpy.ndarray): An integer index mask for selecting a subset of the images to plot.
            subplots_kwargs (dict): Keyword arguments to pass to the figure generation. (Default is None.)
            imshow_kwargs (dict): Keyword arguments to pass to the `imshow` plotting command. (Default is None.)
            hide_axes (bool): Whether to hide axis ticks and labels. (Default is True.)

        Returns:
            (matplotlib.figure.Figure): The figure the plots are in.
            (list): The axes the plot is on.
        """

        if mask is not None:
            images = self.images[mask]
        else:
            images = self.images

        subplots_kwargs = subplots_kwargs or {}
        imshow_kwargs = imshow_kwargs or {}
        nrows, ncols = self._get_factors(len(images))
        fig, axes = plt.subplots(nrows=nrows, ncols=ncols, **subplots_kwargs)
        axes = np.atleast_2d(axes)
        for n, img in enumerate(images):
            i = int(np.floor(n / ncols))
            j = n % ncols
            ax = axes[i, j]
            img.plot(ax=ax, imshow_kwargs=imshow_kwargs, hide_axes=hide_axes)

        fig.tight_layout()
        return fig, axes

    def add_image(self, source, metadata=None, as_gray=False, relative_path=True):
        """
        Add an image to the job.

        Args:
            source (str/numpy.ndarray): The filepath to the data, or the raw array of data itself.
            metadata (Metadata): The metadata associated with the source. (Default is None.)
            as_gray (bool): Whether to interpret the new data as grayscale. (Default is False.)
            relative_path (bool): Whether the path provided is relative. (Default is True, automatically converts to an
                absolute path before setting the `source` value of the image.)
        """

        if not isfile(source) and not isinstance(source, np.ndarray):
            raise ValueError("Could not find a file at {}, nor is source an array.".format(source))
        if isinstance(source, str) and relative_path:
            source = abspath(source)
        self.images.append(Image(source=source, metadata=metadata, as_gray=as_gray))

    def add_images(self, sources, metadata=None, as_gray=False):
        """
        Add multiple images to the job.

        Args:
            sources (str/list/tuple/numpy.ndarray): When a string, uses the `glob` module to look for matching files.
                When list-like, iteratively uses each element as a new source.
            metadata (Metadata): The metadata associated with all these sources. (Default is None.)
            as_gray (bool): Whether to interpret all this data as grayscale. (Default is False.)
            relative_path (bool): Whether the path provided is relative. (Default is True, automatically converts to an
                absolute path before setting the `source` value of the image.)
        """

        if isinstance(sources, str):
            for match in iglob(sources):
                self.add_image(match, metadata=metadata, as_gray=as_gray)
        elif isinstance(sources, (list, tuple, np.ndarray)):
            for source in sources:
                self.add_image(source, metadata=metadata, as_gray=as_gray)

    def run(self, run_again=False, repair=False, debug=False, run_mode=None):
        super(ImageJob, self).run(run_again=run_again, repair=repair, debug=debug, run_mode=run_mode)

    def run_static(self):
        """This is just a toy example right now."""
        self.status.running = True
        if hasattr(self.input, 'filter') and self.input.filter == 'brightness_filter':
            fractions = []
            cutoffs = []
            masks = []
            for img in self.images:
                frac, cut, mask = brightness_filter(img)
                fractions.append(frac)
                cutoffs.append(cut)
                masks.append(mask)
            self.output.fractions = np.array(fractions)
            self.output.cutoffs = np.array(cutoffs)
            self.output.masks = np.array(masks)

        else:
            self.logger.warning("Didn't run anything. Check input.")
        self.status.collect = True
        self.run()

    def write_input(self):
        """Must define abstract method"""
        pass

    def collect_output(self):
        """Must define abstract method"""
        self.to_hdf()

    def to_hdf(self, hdf=None, group_name=None):
        """
        Store the job in an HDF5 file.

        Args:
            hdf (ProjectHDFio): HDF5 group object - optional
            group_name (str): HDF5 subgroup name - optional
        """
        super(ImageJob, self).to_hdf(hdf=hdf, group_name=group_name)
        self.input.to_hdf(hdf=self._hdf5, group_name=None)
        self.output.to_hdf(hdf=self._hdf5, group_name=None)
        with self._hdf5.open("images") as hdf5_server:
            for n, image in enumerate(self.images):
                image.to_hdf(hdf=hdf5_server, group_name="img{}".format(n))
        self._hdf5["n_images"] = n + 1

    def from_hdf(self, hdf=None, group_name=None):
        """
        Load the Protocol from an HDF5 file.

        Args:
            hdf (ProjectHDFio): HDF5 group object - optional
            group_name (str): HDF5 subgroup name - optional
        """
        super(ImageJob, self).from_hdf(hdf=hdf, group_name=group_name)
        self.input.from_hdf(hdf=self._hdf5, group_name=None)
        self.output.from_hdf(hdf=self._hdf5, group_name=None)
        with self._hdf5.open("images") as hdf5_server:
            for n in np.arange(self._hdf5["n_images"], dtype=int):
                img = Image(source=None)
                img.from_hdf(hdf=hdf5_server, group_name="img{}".format(n))
                self.images.append(img)




from pyiron_base import GenericMaster, state
from pyiron_gpl.elastic.elastic import ElasticMatrixJob


class ElasticMatrixJobWithoutFiles(ElasticMatrixJob):
    def __init__(self, project, job_name):
        if not state.database.database_is_disabled:
            raise RuntimeError(
                "To run a `Without` job, the database must first be disabled. Please "
                "`from pyiron_base import state; "
                "state.update({'disable_database': True})`, and try again."
            )
        super(ElasticMatrixJobWithoutFiles, self).__init__(project, job_name)
        self._interactive_disable_log_file = True

    @property
    def child_project(self):
        """
        :class:`.Project`: project which holds the created child jobs
        """
        if not self._interactive_disable_log_file:
            return super(ElasticMatrixJobWithoutFiles, self).child_project
        else:
            return self.project

    def to_hdf(self, hdf=None, group_name=None):
        """

        Args:
            hdf:
            group_name:

        Returns:

        """
        if not self._interactive_disable_log_file:
            super(ElasticMatrixJobWithoutFiles, self).to_hdf(
                hdf=hdf, group_name=group_name
            )

    def refresh_job_status(self):
        if not self._interactive_disable_log_file:
            super(ElasticMatrixJobWithoutFiles).refresh_job_status()

    def collect_output(self):
        if not self._data:
            self.from_hdf()
        self.create_calculator()

        energies = {}
        self._data["id"] = []
        if self.server.run_mode.interactive and not self._interactive_disable_log_file:
            child_id = self.child_ids[0]
            self._data["id"].append(child_id)
            child_job = self.project_hdf5.inspect(child_id)
            energies = {
                job_name: energy
                for job_name, energy in zip(
                    self.structure_dict.keys(), child_job["output/generic/energy_tot"]
                )
            }
        elif self.server.run_mode.interactive and self._interactive_disable_log_file:
            energies = {
                job_name: energy
                for job_name, energy in zip(
                    self.structure_dict.keys(),
                    self.ref_job.interactive_cache["energy_tot"],
                )
            }
        else:
            for job_id in self.child_ids:
                ham = self.project_hdf5.inspect(job_id)
                en = ham["output/generic/energy_tot"][-1]
                energies[ham.job_name] = en
                self._data["id"].append(ham.job_id)

        self.property_calculator.analyse_structures(energies)
        self._data.update(self.property_calculator._data)
        if not self._interactive_disable_log_file:
            self.to_hdf()

    def append(self, job):
        """
        Append a job to the GenericMaster - just like you would append an element to a list.

        Args:
            job (GenericJob): job to append
        """
        if self.status.initialized and not job.status.initialized:
            raise ValueError(
                "GenericMaster requires reference jobs to have status initialized, rather than ",
                job.status.string,
            )
        if job.server.cores >= self.server.cores:
            self.server.cores = job.server.cores
        if job.job_name not in self._job_name_lst:
            self._job_name_lst.append(job.job_name)
            if not self._interactive_disable_log_file:
                self._child_job_update_hdf(parent_job=self, child_job=job)

    def pop(self, i=-1):
        """
        Pop a job from the GenericMaster - just like you would pop an element from a list

        Args:
            i (int): position of the job. (Default is last element, -1.)

        Returns:
            GenericJob: job
        """
        job_name_to_return = self._job_name_lst[i]
        job_to_return = self._load_all_child_jobs(
            self._load_job_from_cache(job_name_to_return)
        )
        del self._job_name_lst[i]
        if not self._interactive_disable_log_file:
            with self.project_hdf5.open("input") as hdf5_input:
                hdf5_input["job_list"] = self._job_name_lst
            job_to_return.relocate_hdf5()
        if isinstance(job_to_return, GenericMaster):
            for sub_job in job_to_return._job_object_dict.values():
                self._child_job_update_hdf(parent_job=job_to_return, child_job=sub_job)
        job_to_return.status.initialized = True
        return job_to_return

    def _run_if_collect(self):
        """
        Internal helper function the run if collect function is called when the job status is 'collect'. It collects
        the simulation output using the standardized functions collect_output() and collect_logfiles(). Afterwards the
        status is set to 'finished'.
        """

        if not self._interactive_disable_log_file:
            super(ElasticMatrixJobWithoutFiles)._run_if_collect()
        else:
            self._logger.info(
                "{}, status: {}, finished".format(self.job_info_str, self.status)
            )
            self.collect_output()
            self._logger.info(
                "{}, status: {}, parallel master".format(self.job_info_str, self.status)
            )
            self.update_master()
            # self.send_to_database()



from pyiron_base import GenericMaster, state
from pyiron_atomistics.atomistics.master.phonopy import PhonopyJob


class PhonopyJobWithoutFiles(PhonopyJob):
    def __init__(self, project, job_name):
        if not state.database.database_is_disabled:
            raise RuntimeError(
                "To run a `Without` job, the database must first be disabled. Please "
                "`from pyiron_base import state; "
                "state.update({'disable_database': True})`, and try again."
            )
        super(PhonopyJobWithoutFiles, self).__init__(project, job_name)
        self._interactive_disable_log_file = True

    @property
    def child_project(self):
        """
        :class:`.Project`: project which holds the created child jobs
        """
        if not self._interactive_disable_log_file:
            return super(PhonopyJobWithoutFiles, self).child_project
        else:
            return self.project

    def to_hdf(self, hdf=None, group_name=None):
        """
        Args:
            hdf:
            group_name:
        Returns:
        """
        if not self._interactive_disable_log_file:
            super(PhonopyJobWithoutFiles, self).to_hdf(hdf=hdf, group_name=group_name)

    def refresh_job_status(self):
        if not self._interactive_disable_log_file:
            super(PhonopyJobWithoutFiles).refresh_job_status()

    def _run_if_collect(self):
        """
        Internal helper function the run if collect function is called when the job status is 'collect'. It collects
        the simulation output using the standardized functions collect_output() and collect_logfiles(). Afterwards the
        status is set to 'finished'.
        """

        if not self._interactive_disable_log_file:
            super(PhonopyJobWithoutFiles)._run_if_collect()
        else:
            self._logger.info(
                "{}, status: {}, finished".format(self.job_info_str, self.status)
            )
            self.collect_output()
            self._logger.info(
                "{}, status: {}, parallel master".format(self.job_info_str, self.status)
            )
            self.update_master()
            self.status.finished = True
            # self.send_to_database()

    def collect_output(self):
        """
        Returns:
        """
        if (
            self.ref_job.server.run_mode.interactive
            and self._interactive_disable_log_file
        ):
            forces_lst = self.ref_job.output.forces
        elif self.ref_job.server.run_mode.interactive:
            forces_lst = self.project_hdf5.inspect(self.child_ids[0])[
                "output/generic/forces"
            ]
        else:
            pr_job = self.project_hdf5.project.open(self.job_name + "_hdf5")
            forces_lst = [
                pr_job.inspect(job_name)["output/generic/forces"][-1]
                for job_name in self._get_jobs_sorted()
            ]
        self.phonopy.set_forces(forces_lst)
        self.phonopy.produce_force_constants(
            fc_calculator=None if self.input["number_of_snapshots"] is None else "alm"
        )
        self.phonopy.run_mesh(mesh=[self.input["dos_mesh"]] * 3)
        mesh_dict = self.phonopy.get_mesh_dict()
        self.phonopy.run_total_dos()
        dos_dict = self.phonopy.get_total_dos_dict()

        if not self._interactive_disable_log_file:
            self.to_hdf()
            with self.project_hdf5.open("output") as hdf5_out:
                hdf5_out["dos_total"] = dos_dict["total_dos"]
                hdf5_out["dos_energies"] = dos_dict["frequency_points"]
                hdf5_out["qpoints"] = mesh_dict["qpoints"]
                hdf5_out["supercell_matrix"] = self._phonopy_supercell_matrix()
                hdf5_out[
                    "displacement_dataset"
                ] = self.phonopy.get_displacement_dataset()
                hdf5_out[
                    "dynamical_matrix"
                ] = self.phonopy.dynamical_matrix.get_dynamical_matrix()
                hdf5_out["force_constants"] = self.phonopy.force_constants

    def append(self, job):
        """
        Append a job to the GenericMaster - just like you would append an element to a list.
        Args:
            job (GenericJob): job to append
        """
        if self.status.initialized and not job.status.initialized:
            raise ValueError(
                "GenericMaster requires reference jobs to have status initialized, rather than ",
                job.status.string,
            )
        if job.server.cores >= self.server.cores:
            self.server.cores = job.server.cores
        if job.job_name not in self._job_name_lst:
            self._job_name_lst.append(job.job_name)
            if not self._interactive_disable_log_file:
                self._child_job_update_hdf(parent_job=self, child_job=job)

    def pop(self, i=-1):
        """
        Pop a job from the GenericMaster - just like you would pop an element from a list
        Args:
            i (int): position of the job. (Default is last element, -1.)
        Returns:
            GenericJob: job
        """
        job_name_to_return = self._job_name_lst[i]
        job_to_return = self._load_all_child_jobs(
            self._load_job_from_cache(job_name_to_return)
        )
        del self._job_name_lst[i]
        if not self._interactive_disable_log_file:
            with self.project_hdf5.open("input") as hdf5_input:
                hdf5_input["job_list"] = self._job_name_lst
            job_to_return.relocate_hdf5()
        if isinstance(job_to_return, GenericMaster):
            for sub_job in job_to_return._job_object_dict.values():
                self._child_job_update_hdf(parent_job=job_to_return, child_job=sub_job)
        job_to_return.status.initialized = True
        return job_to_return

    def plot_dos(self, ax=None, *args, axis=None, **kwargs):
        """
        Plot the DOS.
        If "label" is present in `kwargs` a legend is added to the plot automatically.
        Args:
            axis (optional): matplotlib axis to use, if None create a new one
            ax: deprecated alias for axis
            *args: passed to `axis.plot`
            **kwargs: passed to `axis.plot`
        Returns:
            matplotlib.axes._subplots.AxesSubplot: axis with the plot
        """
        try:
            import pylab as plt
        except ImportError:
            import matplotlib.pyplot as plt
        if ax is not None and axis is None:
            axis = ax
        if axis is None:
            _, axis = plt.subplots(1, 1)
        if not self._interactive_disable_log_file:
            axis.plot(
                self["output/dos_energies"], self["output/dos_total"], *args, **kwargs
            )
        else:
            dos_dict = self.phonopy.get_total_dos_dict()
            axis.plot(
                dos_dict["frequency_points"], dos_dict["total_dos"], *args, **kwargs
            )
        axis.set_xlabel("Frequency [THz]")
        axis.set_ylabel("DOS")
        axis.set_title("Phonon DOS vs Energy")
        if "label" in kwargs:
            axis.legend()
        return ax



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

from __future__ import print_function

from pyiron_contrib.protocol.generic import CompoundVertex, Protocol
from pyiron_contrib.protocol.primitive.one_state import Counter, ExternalHamiltonian, GradientDescent, Max, Norm
from pyiron_contrib.protocol.primitive.two_state import IsGEq
from pyiron_contrib.protocol.utils import Pointer

"""
Protocol for minimizing forces.
"""

__author__ = "Liam Huber, Jan Janssen"
__copyright__ = "Copyright 2019, Max-Planck-Institut für Eisenforschung GmbH " \
                "- Computational Materials Design (CM) Department"
__version__ = "0.0"
__maintainer__ = "Liam Huber"
__email__ = "huber@mpie.de"
__status__ = "development"
__date__ = "May 20, 2019"


class Minimize(CompoundVertex):
    """
    Run minimization with Lammps. This isn't physically useful, since a regular lammps job is faster it's just a dummy
    class for debugging new code and teaching ideas.

    Input attributes:
        ref_job_full_path (str): Path to the pyiron job to use for evaluating forces and energies.
        structure (Atoms): The structure to minimize.
        n_steps (int): How many steps to run for. (Default is 100.)
        f_tol (float): Ionic force convergence (largest atomic force). (Default is 1e-4 eV/angstrom.)
        gamma0 (float): Initial step size as a multiple of the force. (Default is 0.1.)
        fix_com (bool): Whether the center of mass motion should be subtracted off of the position update. (Default is
            True)
        use_adagrad (bool): Whether to have the step size decay according to adagrad. (Default is False)

    Output attributes:
        energy_pot (float): Total potential energy of the system in eV.
        max_force (float): The largest atomic force magnitude in eV/angstrom.
        positions (numpy.ndarray): Atomic positions in angstroms.
        forces (numpy.ndarray): Atomic forces in eV/angstrom. Note: These are the potential gradient forces; thermostat
            forces (if any) are not saved.
    """

    DefaultWhitelist = {
        'calc_static': {
            'output': {
                'energy_pot': 1
            }
        },
        'max_force': {
            'output': {
                'amax': 1
            }
        }
    }

    def __init__(self, **kwargs):
        super(Minimize, self).__init__(**kwargs)

        # Protocol defaults
        id_ = self.input.default
        id_.n_steps = 100
        id_.f_tol = 1e-4
        id_.gamma0 = 0.1
        id_.fix_com = True
        id_.use_adagrad = False

    def define_vertices(self):
        # Graph components
        g = self.graph
        g.calc_static = ExternalHamiltonian()
        g.clock = Counter()
        g.check_steps = IsGEq()
        g.force_norm = Norm()
        g.max_force = Max()
        g.check_force = IsGEq()
        g.gradient_descent = GradientDescent()

    def define_execution_flow(self):
        # Execution flow
        g = self.graph
        g.make_pipeline(
            g.check_steps, 'false',
            g.calc_static,
            g.force_norm,
            g.max_force,
            g.gradient_descent,
            g.check_force, 'true',
            g.clock,
            g.check_steps
        )
        g.starting_vertex = self.graph.check_steps
        g.restarting_vertex = self.graph.check_steps

    def define_information_flow(self):
        # Data flow
        g = self.graph
        gp = Pointer(self.graph)
        ip = Pointer(self.input)

        # check_steps
        g.check_steps.input.target = gp.clock.output.n_counts[-1]
        g.check_steps.input.threshold = ip.n_steps

        # calc_static
        g.calc_static.input.ref_job_full_path = ip.ref_job_full_path
        g.calc_static.input.structure = ip.structure
        g.calc_static.input.default.positions = ip.structure.positions
        g.calc_static.input.positions = gp.gradient_descent.output.positions[-1]

        # force_norm
        g.force_norm.input.x = gp.calc_static.output.forces[-1]
        g.force_norm.input.ord = 2
        g.force_norm.input.axis = -1

        # max_force
        g.max_force.input.a = gp.force_norm.output.n[-1]

        # gradient_descent
        g.gradient_descent.input.default.positions = ip.structure.positions
        g.gradient_descent.input.positions = gp.gradient_descent.output.positions[-1]
        g.gradient_descent.input.forces = gp.calc_static.output.forces[-1]
        g.gradient_descent.input.masses = ip.structure.get_masses
        g.gradient_descent.input.gamma0 = ip.gamma0
        g.gradient_descent.input.fix_com = ip.fix_com
        g.gradient_descent.input.use_adagrad = ip.use_adagrad

        # check_force
        g.check_force.input.target = gp.max_force.output.amax[-1]
        g.check_force.input.threshold = ip.f_tol

        self.set_graph_archive_clock(gp.clock.output.n_counts[-1])

    def get_output(self):
        gp = Pointer(self.graph)
        return {
            'energy_pot': ~gp.calc_static.output.energy_pot[-1],
            'max_force': ~gp.max_force.output.amax[-1],
            'positions': ~gp.gradient_descent.output.positions[-1],
            'forces': ~gp.calc_static.output.forces[-1]
        }


class ProtoMinimGradDes(Protocol, Minimize):
    pass



# coding: utf-8
# Copyright (c) ICAMS, Ruhr University Bochum, 2022

## Executable required: $pyiron/resources/pacemaker/bin/run_pacemaker_tf_cpu.sh AND  run_pacemaker_tf.sh

import logging
from typing import List

import numpy as np
import os
import pandas as pd
import ruamel.yaml as yaml

from shutil import copyfile

from pyiron_base import GenericJob, GenericParameters, state, Executable, FlattenedStorage

from pyiron_contrib.atomistics.atomistics.job.trainingcontainer import TrainingStorage, TrainingContainer
from pyiron_contrib.atomistics.ml.potentialfit import PotentialFit

from pyiron_atomistics.atomistics.structure.atoms import Atoms as pyironAtoms, ase_to_pyiron
from ase.atoms import Atoms as aseAtoms

s = state.settings


class PacemakerJob(GenericJob, PotentialFit):

    def __init__(self, project, job_name):
        super().__init__(project, job_name)
        self.__name__ = "Pacemaker2022"
        self.__version__ = "0.2"

        self._train_job_id_list = []

        self.input = GenericParameters(table_name="input")
        self._cutoff = 7.0
        self.input['cutoff'] = self._cutoff
        self.input['metadata'] = {'comment': 'pyiron-generated fitting job'}

        # data_config
        self.input['data'] = {}
        # potential_config
        self.input['potential'] = {
            "elements": [],
            "bonds": {
                "ALL": {
                    "radbase": "SBessel",
                    "rcut": self._cutoff,
                    "dcut": 0.01,
                    "radparameters": [5.25]
                }
            },

            "embeddings": {
                "ALL": {
                    "fs_parameters": [1, 1, 1, 0.5],
                    "ndensity": 2,
                    "npot": "FinnisSinclairShiftedScaled"
                }
            },

            "functions": {
                "ALL": {
                    "nradmax_by_orders": [15, 3, 2, 1],
                    "lmax_by_orders": [0, 3, 2, 1],
                }
            }
        }

        # fit_config
        self.input['fit'] = {
            "loss": {"L1_coeffs": 1e-8, "L2_coeffs": 1e-8, "kappa": 0.3, "w0_rad": 0,
                     "w1_rad": 0, "w2_rad": 0},
            "maxiter": 1000,
            "optimizer": "BFGS",
            "fit_cycles": 1
        }
        self.input['backend'] = {"batch_size": 100,
                                 "display_step": 50,
                                 "evaluator": "tensorpot"}  # backend_config

        self.structure_data = None
        self._executable = None
        self._executable_activate()

        state.publications.add(self.publication)

    @property
    def elements(self):
        return self.input["potential"].get("elements")

    @elements.setter
    def elements(self, val):
        self.input["potential"]["elements"] = val

    @property
    def cutoff(self):
        return self._cutoff

    @cutoff.setter
    def cutoff(self, val):
        self._cutoff = val
        self.input["cutoff"] = self._cutoff
        self.input["potential"]["bonds"]["ALL"]["rcut"] = self._cutoff

    @property
    def publication(self):
        return {
            "pacemaker": [
                {
                    "title": "Efficient parametrization of the atomic cluster expansion",
                    "journal": "Physical Review Materials",
                    "volume": "6",
                    "number": "1",
                    "year": "2022",
                    "doi": "10.1103/PhysRevMaterials.6.013804",
                    "url": "https://doi.org/10.1103/PhysRevMaterials.6.013804",
                    "author": ["Anton Bochkarev", "Yury Lysogorskiy", "Sarath Menon", "Minaam Qamar", "Matous Mrovec",
                               "Ralf Drautz"],
                },

                {
                    "title": "Performant implementation of the atomic cluster expansion (PACE) and application to copper and silicon",
                    "journal": "npj Computational Materials",
                    "volume": "7",
                    "number": "1",
                    "year": "2021",
                    "doi": "10.1038/s41524-021-00559-9",
                    "url": "https://doi.org/10.1038/s41524-021-00559-9",
                    "author": ["Yury Lysogorskiy", "Cas van der Oord", "Anton Bochkarev", "Sarath Menon",
                               "Matteo Rinaldi",
                               "Thomas Hammerschmidt", "Matous Mrovec", "Aidan Thompson", "Gábor Csányi",
                               "Christoph Ortner",
                               "Ralf Drautz"],
                },
                {
                    "title": "Atomic cluster expansion for accurate and transferable interatomic potentials",
                    "journal": "Physical Review B",
                    "volume": "99",
                    "year": "2019",
                    "doi": "10.1103/PhysRevB.99.014104",
                    "url": "https://doi.org/10.1103/PhysRevB.99.014104",
                    "author": ["Ralf Drautz"],
                },
            ]
        }

    def _save_structure_dataframe_pckl_gzip(self, df):

        if "NUMBER_OF_ATOMS" not in df.columns and "number_of_atoms" in df.columns:
            df.rename(columns={"number_of_atoms": "NUMBER_OF_ATOMS"}, inplace=True)
        df["NUMBER_OF_ATOMS"] = df["NUMBER_OF_ATOMS"].astype(int)

        # TODO: reference energy subtraction ?
        if "energy_corrected" not in df.columns and "energy" in df.columns:
            df.rename(columns={"energy": "energy_corrected"}, inplace=True)

        if "atoms" in df.columns:
            # check if this is pyironAtoms  -> aseAtoms
            at = df.iloc[0]["atoms"]
            if isinstance(at, pyironAtoms):
                df["ase_atoms"] = df["atoms"].map(lambda s: s.to_ase())
                df.drop(columns=["atoms"], inplace=True)
            else:
                assert isinstance(at, aseAtoms), "'atoms' column is not a valid ASE Atoms object"
                df.rename(columns={"atoms": "ase_atom"}, inplace=True)
        elif "ase_atoms" not in df.columns:
            raise ValueError("DataFrame should contain 'atoms' (pyiron Atoms) or 'ase_atoms' (ASE atoms) columns")

        if "stress" in df.columns:
            df.drop(columns=["stress"], inplace=True)

        if "pbc" not in df.columns:
            df["pbc"] = df["ase_atoms"].map(lambda atoms: np.all(atoms.pbc))

        data_file_name = os.path.join(self.working_directory, "df_fit.pckl.gzip")
        logging.info("Saving training structures dataframe into {} with pickle protocol = 4, compression = gzip".format(
            data_file_name))
        df.to_pickle(data_file_name, compression="gzip", protocol=4)
        return data_file_name

    def write_input(self):
        # prepare datafile
        if self._train_job_id_list and self.structure_data is None:
            train_df = self.create_training_dataframe(self._train_job_id_list)
            self.structure_data = train_df

        if isinstance(self.structure_data, pd.DataFrame):
            logging.info("structure_data is pandas.DataFrame")
            data_file_name = self._save_structure_dataframe_pckl_gzip(self.structure_data)
            self.input["data"] = {"filename": data_file_name}
            elements_set = set()
            for at in self.structure_data["ase_atoms"]:
                elements_set.update(at.get_chemical_symbols())
            elements = sorted(elements_set)
            print("Set automatically determined list of elements: {}".format(elements))
            self.elements = elements
        elif isinstance(self.structure_data, str):  # filename
            if os.path.isfile(self.structure_data):
                logging.info("structure_data is valid file path")
                self.input["data"] = {"filename": self.structure_data}
            else:
                raise ValueError("Provided structure_data filename ({}) doesn't exists".format(self.structure_data))
        elif hasattr(self.structure_data, "get_pandas"):  # duck-typing check for TrainingContainer
            logging.info("structure_data is TrainingContainer")
            df = self.structure_data.to_pandas()
            data_file_name = self._save_structure_dataframe_pckl_gzip(df)
            self.input["data"] = {"filename": data_file_name}
        elif self.structure_data is None:
            raise ValueError(
                "`structure_data` is none, but should be pd.DataFrame, TrainingContainer or valid pickle.gzip filename")

        metadata_dict = self.input["metadata"]
        metadata_dict["pyiron_job_id"] = str(self.job_id)

        input_yaml_dict = {
            "cutoff": self.input["cutoff"],
            "metadata": metadata_dict,
            "potential": self.input["potential"],
            "data": self.input["data"],
            "fit": self.input["fit"],
            'backend': self.input["backend"],
        }

        if isinstance(self.input["potential"], str):
            pot_file_name = self.input["potential"]
            if os.path.isfile(pot_file_name):
                logging.info("Input potential is filename")
                pot_basename = os.path.basename(pot_file_name)
                copyfile(pot_file_name, os.path.join(self.working_directory, pot_basename))
                input_yaml_dict['potential'] = pot_basename
            else:
                raise ValueError("Provided potential filename ({}) doesn't exists".format(self.input["potential"]))

        with open(os.path.join(self.working_directory, "input.yaml"), "w") as f:
            yaml.dump(input_yaml_dict, f)

    def _analyse_log(self, logfile="metrics.txt"):
        metrics_filename = os.path.join(self.working_directory, logfile)

        metrics_df = pd.read_csv(metrics_filename, sep="\s+")
        res_dict = metrics_df.to_dict(orient="list")
        return res_dict

    def collect_output(self):
        final_potential_filename_yaml = self.get_final_potential_filename()
        with open(final_potential_filename_yaml, "r") as f:
            yaml_lines = f.readlines()
        final_potential_yaml_string = "".join(yaml_lines)

        with open(self.get_final_potential_filename_ace(), "r") as f:
            ace_lines = f.readlines()
        final_potential_yace_string = "".join(ace_lines)

        with open(self.get_final_potential_filename_ace(), "r") as f:
            yace_data = yaml.safe_load(f)

        elements_name = yace_data["elements"]

        with self.project_hdf5.open("output/potential") as h5out:
            h5out["yaml"] = final_potential_yaml_string
            h5out["yace"] = final_potential_yace_string
            h5out["elements_name"] = elements_name

        log_res_dict = self._analyse_log()

        with self.project_hdf5.open("output/log") as h5out:
            for key, arr in log_res_dict.items():
                h5out[key] = arr

        # training data
        training_data_fname = os.path.join(self.working_directory, "fitting_data_info.pckl.gzip")
        df = pd.read_pickle(training_data_fname, compression="gzip")
        df["atoms"] = df.ase_atoms.map(ase_to_pyiron)
        training_data_ts = TrainingStorage()
        for _, r in df.iterrows():
            training_data_ts.add_structure(r.atoms,
                                           energy=r.energy_corrected,
                                           forces=r.forces,
                                           identifier=r['name'])

        # predicted data
        predicted_fname = os.path.join(self.working_directory, "train_pred.pckl.gzip")
        df = pd.read_pickle(predicted_fname, compression="gzip")
        predicted_data_fs = FlattenedStorage()
        predicted_data_fs.add_array('energy', dtype=np.float64, shape=(), per='chunk')
        predicted_data_fs.add_array('energy_true', dtype=np.float64, shape=(), per='chunk')

        predicted_data_fs.add_array('number_of_atoms', dtype=np.int64, shape=(), per='chunk')

        predicted_data_fs.add_array('forces', dtype=np.float64, shape=(3,), per='element')
        predicted_data_fs.add_array('forces_true', dtype=np.float64, shape=(3,), per='element')
        for i, r in df.iterrows():
            identifier = r['name'] if "name" in r else str(i)
            predicted_data_fs.add_chunk(r["NUMBER_OF_ATOMS"], identifier=identifier,
                                        energy=r.energy_pred,
                                        forces=r.forces_pred,
                                        energy_true=r.energy_corrected,
                                        forces_true=r.forces,
                                        number_of_atoms = r.NUMBER_OF_ATOMS,

                                        energy_per_atom = r.energy_pred / r.NUMBER_OF_ATOMS,
                                        energy_per_atom_true=r.energy_corrected / r.NUMBER_OF_ATOMS,
                                        )

        with self.project_hdf5.open("output") as hdf5_output:
            training_data_ts.to_hdf(hdf=hdf5_output, group_name="training_data")
            predicted_data_fs.to_hdf(hdf=hdf5_output, group_name="predicted_data")

    def get_lammps_potential(self):
        elements_name = self["output/potential/elements_name"]
        elem = " ".join(elements_name)
        pot_file_name = self.get_final_potential_filename_ace()
        pot_dict = {
            'Config': [["pair_style pace\n", "pair_coeff  * * {} {}\n".format(pot_file_name, elem)]],
            'Filename': [""],
            'Model': ["ACE"],
            'Name': [self.job_name],
            'Species': [elements_name]
        }

        ace_potential = pd.DataFrame(pot_dict)

        return ace_potential

    def to_hdf(self, hdf=None, group_name=None):
        super().to_hdf(
            hdf=hdf,
            group_name=group_name
        )
        with self.project_hdf5.open("input") as h5in:
            self.input.to_hdf(h5in)

    def from_hdf(self, hdf=None, group_name=None):
        super().from_hdf(
            hdf=hdf,
            group_name=group_name
        )
        with self.project_hdf5.open("input") as h5in:
            self.input.from_hdf(h5in)

    def get_final_potential_filename(self):
        return os.path.join(self.working_directory, "output_potential.yaml")

    def get_final_potential_filename_ace(self):
        return os.path.join(self.working_directory, "output_potential.yace")

    def get_current_potential_filename(self):
        return os.path.join(self.working_directory, "interim_potential_0.yaml")

    # To link to the executable from the notebook
    def _executable_activate(self, enforce=False, codename="pacemaker"):
        if self._executable is None or enforce:
            self._executable = Executable(
                codename=codename, module="pacemaker", path_binary_codes=state.settings.resource_paths
            )

    def _add_training_data(self, container: TrainingContainer) -> None:
        self.add_job_to_fitting(container.id)

    def add_job_to_fitting(self, job_id, *args, **kwargs):
        self._train_job_id_list.append(job_id)

    def _get_training_data(self) -> TrainingStorage:
        return self["output/training_data"].to_object()

    def _get_predicted_data(self) -> FlattenedStorage:
        return self["output/predicted_data"].to_object()

    # copied/adapted from mlip.py
    def create_training_dataframe(self, _train_job_id_list: List = None) -> pd.DataFrame:
        if _train_job_id_list is None:
            _train_job_id_list = self._train_job_id_list
        df_list = []
        for job_id in _train_job_id_list:
            ham = self.project.inspect(job_id)
            if ham.__name__ == "TrainingContainer":
                job = ham.to_object()
                data_df = job.to_pandas()
                df_list.append(data_df)
            else:
                raise NotImplementedError("Currently only TrainingContainer is supported")

        total_training_df = pd.concat(df_list, axis=0)
        total_training_df.reset_index(drop=True, inplace=True)

        return total_training_df



__version__ = "0.1"
__all__ = []

import warnings

try:
	from pyiron import Project
except:
	warnings.warn("pyiron module not found, importing Project from pyiron_base")
	from pyiron_base import Project

from pyiron_base import JOB_CLASS_DICT
from pyiron_contrib.generic.storage_interface_toolkit import StorageInterfaceFactory

Project.register_tools('storage_interface', StorageInterfaceFactory)

# Make classes available for new pyiron version
JOB_CLASS_DICT['ProtoMinimGradDes'] = 'pyiron_contrib.protocol.compound.minimize'
JOB_CLASS_DICT['ProtoMD'] = 'pyiron_contrib.protocol.compound.molecular_dynamics'
JOB_CLASS_DICT['ProtoConfinedMD'] = 'pyiron_contrib.protocol.compound.molecular_dynamics'
JOB_CLASS_DICT['ProtoHarmMD'] = 'pyiron_contrib.protocol.compound.molecular_dynamics'
JOB_CLASS_DICT['ProtoNEBSer'] = 'pyiron_contrib.protocol.compound.nudged_elastic_band'
JOB_CLASS_DICT['ProtocolQMMM'] = 'pyiron_contrib.protocol.compound.qmmm'
JOB_CLASS_DICT['ProtoHarmTILDSer'] = 'pyiron_contrib.protocol.compound.thermodynamic_integration'
JOB_CLASS_DICT['ProtoHarmTILDPar'] = 'pyiron_contrib.protocol.compound.thermodynamic_integration'
JOB_CLASS_DICT['ProtoVacTILDSer'] = 'pyiron_contrib.protocol.compound.thermodynamic_integration'
JOB_CLASS_DICT['ProtoVacTILDPar'] = 'pyiron_contrib.protocol.compound.thermodynamic_integration'
JOB_CLASS_DICT['ProtoVacForm'] = 'pyiron_contrib.protocol.compound.thermodynamic_integration'
JOB_CLASS_DICT['ProtoFTSEvoSer'] = 'pyiron_contrib.protocol.compound.finite_temperature_string'
JOB_CLASS_DICT['ProtoFTSEvoPar'] = 'pyiron_contrib.protocol.compound.finite_temperature_string'
JOB_CLASS_DICT['ImageJob'] = 'pyiron_contrib.image.job'
JOB_CLASS_DICT['LangevinAse'] = 'pyiron_contrib.atomistics.interactive.langevin'
JOB_CLASS_DICT['Mixer'] = 'pyiron_contrib.atomistics.interactive.mixer'
JOB_CLASS_DICT['ParameterMaster'] = 'pyiron_contrib.atomistics.dft.parametermaster'
JOB_CLASS_DICT['MonteCarloMaster'] = 'pyiron_contrib.atomistics.interactive.montecarlo'
JOB_CLASS_DICT['RandSpg'] = 'pyiron_contrib.atomistics.randspg.randspg'
JOB_CLASS_DICT['Fenics'] = 'pyiron_contrib.continuum.fenics.job.generic'
JOB_CLASS_DICT['FenicsLinearElastic'] = 'pyiron_contrib.continuum.fenics.job.elastic'
JOB_CLASS_DICT['TrainingContainer'] = 'pyiron_contrib.atomistics.atomistics.job.trainingcontainer'
JOB_CLASS_DICT['RandomDisMaster'] = 'pyiron_contrib.atomistics.mlip.masters'
JOB_CLASS_DICT['RandomMDMaster'] = 'pyiron_contrib.atomistics.mlip.masters'
JOB_CLASS_DICT['RunnerFit'] = 'pyiron_contrib.atomistics.runner.job'
JOB_CLASS_DICT['MlipSelect'] = 'pyiron_contrib.atomistics.mlip.mlipselect'
JOB_CLASS_DICT['Mlip'] = 'pyiron_contrib.atomistics.mlip.mlip'
JOB_CLASS_DICT['LammpsMlip'] = 'pyiron_contrib.atomistics.mlip.lammps'
JOB_CLASS_DICT['MlipJob'] = 'pyiron_contrib.atomistics.mlip.mlipjob'
JOB_CLASS_DICT['Atomicrex'] = 'pyiron_contrib.atomistics.atomicrex.atomicrex_job'
JOB_CLASS_DICT['StructureMasterInt'] = 'pyiron_contrib.atomistics.atomistics.job.structurelistmasterinteractive'
JOB_CLASS_DICT['StorageJob'] = 'pyiron_contrib.RDM.storagejob'
JOB_CLASS_DICT['MlipDescriptors'] = 'pyiron_contrib.atomistics.mlip.mlipdescriptors'
JOB_CLASS_DICT['PacemakerJob'] = 'pyiron_contrib.atomistics.pacemaker.job'
JOB_CLASS_DICT['MeamFit'] = 'pyiron_contrib.atomistics.meamfit.meamfit'
JOB_CLASS_DICT['Cp2kJob'] = 'pyiron_contrib.atomistics.cp2k.job'
JOB_CLASS_DICT['PiMD'] = 'pyiron_contrib.atomistics.ipi.ipi_jobs'
JOB_CLASS_DICT['GleMD'] = 'pyiron_contrib.atomistics.ipi.ipi_jobs'
JOB_CLASS_DICT['PigletMD'] = 'pyiron_contrib.atomistics.ipi.ipi_jobs'
JOB_CLASS_DICT['LammpsInteractiveWithoutOutput'] = 'pyiron_contrib.nofiles.lammps'
JOB_CLASS_DICT['SQSJobWithoutOutput'] = 'pyiron_contrib.nofiles.sqs'
JOB_CLASS_DICT['ElasticMatrixJobWithoutFiles'] = 'pyiron_contrib.nofiles.elastic'
JOB_CLASS_DICT['MurnaghanWithoutFiles'] = 'pyiron_contrib.nofiles.murn'
JOB_CLASS_DICT['PhonopyJobWithoutFiles'] = 'pyiron_contrib.nofiles.phonopy'
JOB_CLASS_DICT['SQSMasterMPI'] = 'pyiron_contrib.nofiles.master'
JOB_CLASS_DICT['LAMMPSMinimizeMPI'] = 'pyiron_contrib.nofiles.master'
JOB_CLASS_DICT['LAMMPSElasticMPI'] = 'pyiron_contrib.nofiles.master'
JOB_CLASS_DICT['LAMMPSMinimizeElasticMPI'] = 'pyiron_contrib.nofiles.master'


from ._version import get_versions

__version__ = get_versions()["version"]
del get_versions



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

from __future__ import print_function

from pyiron_contrib.protocol.generic import CompoundVertex, Protocol
from pyiron_contrib.protocol.primitive.one_state import BerendsenBarostat, Counter, CutoffDistance, \
    ExternalHamiltonian, HarmonicHamiltonian, RandomVelocity, SphereReflection, VerletPositionUpdate, \
    VerletVelocityUpdate, Zeros
from pyiron_contrib.protocol.primitive.two_state import IsGEq
from pyiron_contrib.protocol.utils import Pointer

"""
Protocols for molecular dynamics.
"""

__author__ = "Liam Huber"
__copyright__ = "Copyright 2019, Max-Planck-Institut für Eisenforschung GmbH " \
                "- Computational Materials Design (CM) Department"
__version__ = "0.0"
__maintainer__ = "Liam Huber"
__email__ = "huber@mpie.de"
__status__ = "development"
__date__ = "18 July, 2019"


class MolecularDynamics(CompoundVertex):
    """
    Runs molecular dynamics. This isn't particularly useful as almost every source code/plain job can do this on its
     own, but rather this is intended for testing and teaching. It also serves as a useful starting point for
     developing algorithms with modified dynamics.

    Input attributes:
        ref_job_full_path (str): Path to the pyiron job to use for evaluating forces and energies.
        structure (Atoms): The structure evolve.
        temperature (float): Temperature to run at in K.
        n_steps (int): How many MD steps to run for. (Default is 100.)
        temperature_damping_timescale (float): Langevin thermostat timescale in fs. (Default is None, which runs NVE.)
        time_step (float): MD time step in fs. (Default is 1.)
        overheat_fraction (float): The fraction by which to overheat the initial velocities. This can be useful for
            more quickly equilibrating a system whose initial structure is its fully relaxed positions -- in which
            case equipartition of energy tells us that the kinetic energy should be initialized to double the
            desired value. (Default is 2.0, assume energy equipartition is a good idea.)
        pressure (float): The pressure in GPa to be simulated (Default is None GPa)
        pressure_style (string): 'isotropic' or 'anisotropic'. (Default is 'anisotropic')
        pressure_damping_timescale (float): Damping timescale in fs. (Default is None, no barostat is used.)
        compressibility (float): The compressibility of water in bar-1 (Default is 4.57e-5 bar-1)
        previous_volume (float): The default volume. (Defaults is None.)
        energy_kin (float): The default energy_kin. (Default is None.)

    Output attributes:
        energy_pot (float): Total potential energy of the system in eV.
        energy_kin (float): Total kinetic energy of the system in eV.
        positions (numpy.ndarray): Atomic positions in angstroms.
        velocities (numpy.ndarray): Atomic velocities in angstroms/fs.
        forces (numpy.ndarray): Atomic forces in eV/angstrom. Note: These are the potential gradient forces; thermostat
            forces (if any) are not saved.
        pressure (float): The isotropic pressure in GPa.
        volume (float): The volume of the system in Ang3.
    """
    # DefaultWhitelist sets the output which will be stored every `archive period' in the final hdf5 file.

    DefaultWhitelist = {
        'verlet_positions': {
            'output': {
                'positions': 1000,
            },
        },
        'calc_static': {
            'output': {
                'energy_pot': 1,
                'forces': 1000,
            },
        },
        'verlet_velocities': {
            'output': {
                'energy_kin': 1,
                'velocities': 1000,
            },
        },
    }

    def __init__(self, **kwargs):
        super(MolecularDynamics, self).__init__(**kwargs)

        # Protocol defaults
        id_ = self.input.default
        id_.temperature = None
        id_.n_steps = 100
        id_.temperature_damping_timescale = 100.
        id_.time_step = 1.
        id_.overheat_fraction = 2
        id_.pressure = None
        id_.pressure_style = 'anisotropic'
        id_.pressure_damping_timescale = 1000.
        id_._compressibility = 4.57e-5  # bar^-1
        id_._previous_volume = None
        id_._energy_kin = None

    def define_vertices(self):
        # Graph components
        g = self.graph
        g.initial_velocities = RandomVelocity()
        g.initial_forces = Zeros()
        g.initial_pressures = Zeros()
        g.check_steps = IsGEq()
        g.clock = Counter()
        g.barostat = BerendsenBarostat()
        g.verlet_positions = VerletPositionUpdate()
        g.calc_static = ExternalHamiltonian()
        g.verlet_velocities = VerletVelocityUpdate()

    def define_execution_flow(self):
        # Execution flow
        g = self.graph
        g.make_pipeline(
            g.initial_velocities,
            g.initial_forces,
            g.initial_pressures,
            g.check_steps, 'false',
            g.barostat,
            g.verlet_positions,
            g.calc_static,
            g.verlet_velocities,
            g.clock,
            g.check_steps
        )
        g.starting_vertex = g.initial_velocities
        g.restarting_vertex = g.check_steps

    def define_information_flow(self):
        # Data flow
        g = self.graph
        gp = Pointer(self.graph)
        ip = Pointer(self.input)

        # initial_velocities
        g.initial_velocities.input.temperature = ip.temperature
        g.initial_velocities.input.masses = ip.structure.get_masses
        g.initial_velocities.input.overheat_fraction = ip.overheat_fraction

        # initial_forces
        g.initial_forces.input.shape = ip.structure.positions.shape

        # initial_pressures
        g.initial_pressures.input.shape = ip.structure.cell.array.shape

        # check_steps
        g.check_steps.input.target = gp.clock.output.n_counts[-1]
        g.check_steps.input.threshold = ip.n_steps

        # barostat
        g.barostat.input.default.box_pressure = gp.initial_pressures.output.zeros[-1]
        g.barostat.input.default.structure = ip.structure
        g.barostat.input.default.energy_kin = ip._energy_kin
        g.barostat.input.default.previous_volume = ip._previous_volume
        g.barostat.input.default.positions = ip.structure.positions

        g.barostat.input.box_pressure = gp.calc_static.output.pressures[-1]
        g.barostat.input.structure = gp.barostat.output.structure[-1]
        g.barostat.input.energy_kin = gp.verlet_velocities.output.energy_kin[-1]
        g.barostat.input.previous_volume = gp.calc_static.output.volume[-1]
        g.barostat.input.positions = gp.verlet_positions.output.positions[-1]
        g.barostat.input.pressure = ip.pressure
        g.barostat.input.temperature = ip.temperature
        g.barostat.input.time_step = ip.time_step
        g.barostat.input.pressure_damping_timescale = ip.pressure_damping_timescale
        g.barostat.input.compressibility = ip._compressibility
        g.barostat.input.pressure_style = ip.pressure_style

        # verelt_positions
        g.verlet_positions.input.default.velocities = gp.initial_velocities.output.velocities[-1]
        g.verlet_positions.input.default.forces = gp.initial_forces.output.zeros[-1]

        g.verlet_positions.input.positions = gp.barostat.output.positions[-1]
        g.verlet_positions.input.velocities = gp.verlet_velocities.output.velocities[-1]
        g.verlet_positions.input.forces = gp.calc_static.output.forces[-1]
        g.verlet_positions.input.masses = ip.structure.get_masses
        g.verlet_positions.input.time_step = ip.time_step
        g.verlet_positions.input.temperature = ip.temperature
        g.verlet_positions.input.temperature_damping_timescale = ip.temperature_damping_timescale

        # calc_static
        g.calc_static.input.ref_job_full_path = ip.ref_job_full_path
        g.calc_static.input.structure = gp.barostat.output.structure[-1]
        g.calc_static.input.cell = gp.barostat.output.structure[-1].cell.array
        g.calc_static.input.positions = gp.verlet_positions.output.positions[-1]

        # verlet_velocities
        g.verlet_velocities.input.velocities = gp.verlet_positions.output.velocities[-1]
        g.verlet_velocities.input.forces = gp.calc_static.output.forces[-1]
        g.verlet_velocities.input.masses = ip.structure.get_masses
        g.verlet_velocities.input.time_step = ip.time_step
        g.verlet_velocities.input.temperature = ip.temperature
        g.verlet_velocities.input.temperature_damping_timescale = ip.temperature_damping_timescale

        self.set_graph_archive_clock(gp.clock.output.n_counts[-1])

    def get_output(self):
        gp = Pointer(self.graph)
        return {
            'energy_pot': ~gp.calc_static.output.energy_pot[-1],
            'energy_kin': ~gp.verlet_velocities.output.energy_kin[-1],
            'positions': ~gp.verlet_positions.output.positions[-1],
            'velocities': ~gp.verlet_velocities.output.velocities[-1],
            'forces': ~gp.calc_static.output.forces[-1],
            'pressure': ~gp.barostat.output.pressure[-1],
            'volume': ~gp.calc_static.output.volume[-1]
        }


class ProtoMD(Protocol, MolecularDynamics):
    pass


class ConfinedMD(MolecularDynamics):
    """
    Similar to the MolecularDynamics protocol, ConfinedMD performs MD on a structure. The difference, is that the
        atoms are confined to their lattice sites. This is especially helpful when vacancies are present in the
        structure, and atoms diffuse via the vacancies. This protocol prevents this diffusion from happening.

    Input attributes:
        cutoff_factor (float): The cutoff is obtained by taking the first nearest neighbor distance and multiplying
            it by the cutoff factor. A default value of 0.4 is chosen, because taking a cutoff factor of ~0.5
            sometimes let certain reflections off the hook, and we do not want that to happen. (Default is 0.4.)
        use_reflection (boolean): Turn on or off `SphereReflection` (Default is True.)
        total_steps (int): The total number of times `SphereReflection` is called so far. (Default is 0.)

    For inherited input and output attributes, refer the `MolecularDynamics` protocol.
    """

    def __init__(self, **kwargs):
        super(ConfinedMD, self).__init__(**kwargs)

        # Protocol defaults
        id_ = self.input.default
        id_.cutoff_factor = 0.4
        id_.use_reflection = True
        id_.total_steps = 0

    def define_vertices(self):
        # Graph components
        g = self.graph
        g.initial_velocities = RandomVelocity()
        g.initial_forces = Zeros()
        g.initial_pressures = Zeros()
        g.cutoff = CutoffDistance()
        g.check_steps = IsGEq()
        g.clock = Counter()
        g.barostat = BerendsenBarostat()
        g.verlet_positions = VerletPositionUpdate()
        g.reflect_atoms = SphereReflection()
        g.calc_static = ExternalHamiltonian()
        g.verlet_velocities = VerletVelocityUpdate()

    def define_execution_flow(self):
        # Execution flow
        g = self.graph
        g.make_pipeline(
            g.initial_velocities,
            g.initial_forces,
            g.initial_pressures,
            g.cutoff,
            g.check_steps, 'false',
            g.barostat,
            g.verlet_positions,
            g.reflect_atoms,
            g.calc_static,
            g.verlet_velocities,
            g.clock,
            g.check_steps
        )
        g.starting_vertex = g.initial_velocities
        g.restarting_vertex = g.check_steps

    def define_information_flow(self):
        # Data flow
        g = self.graph
        gp = Pointer(self.graph)
        ip = Pointer(self.input)

        # initial_velocities
        g.initial_velocities.input.temperature = ip.temperature
        g.initial_velocities.input.masses = ip.structure.get_masses
        g.initial_velocities.input.overheat_fraction = ip.overheat_fraction

        # initial_forces
        g.initial_forces.input.shape = ip.structure.positions.shape

        # initial_pressures
        g.initial_pressures.input.shape = ip.structure.cell.array.shape

        # cutoff
        g.cutoff.input.structure = ip.structure
        g.cutoff.input.cutoff_factor = ip.cutoff_factor

        # check_steps
        g.check_steps.input.target = gp.clock.output.n_counts[-1]
        g.check_steps.input.threshold = ip.n_steps

        # barostat
        g.barostat.input.default.box_pressure = gp.initial_pressures.output.zeros[-1]
        g.barostat.input.default.structure = ip.structure
        g.barostat.input.default.energy_kin = ip.energy_kin
        g.barostat.input.default.previous_volume = ip.previous_volume
        g.barostat.input.default.positions = ip.structure.positions

        g.barostat.input.box_pressure = gp.calc_static.output.pressures[-1]
        g.barostat.input.structure = gp.barostat.output.structure[-1]
        g.barostat.input.energy_kin = gp.verlet_velocities.output.energy_kin[-1]
        g.barostat.input.previous_volume = gp.calc_static.output.volume[-1]
        g.barostat.input.positions = gp.reflect_atoms.output.positions[-1]
        g.barostat.input.pressure = ip.pressure
        g.barostat.input.temperature = ip.temperature
        g.barostat.input.time_step = ip.time_step
        g.barostat.input.pressure_damping_timescale = ip.pressure_damping_timescale
        g.barostat.input.compressibility = ip.compressibility
        g.barostat.input.pressure_style = ip.pressure_style

        # verlet_positions
        g.verlet_positions.input.default.velocities = gp.initial_velocities.output.velocities[-1]
        g.verlet_positions.input.default.forces = gp.initial_forces.output.zeros[-1]

        g.verlet_positions.input.positions = gp.barostat.output.positions[-1]
        g.verlet_positions.input.velocities = gp.verlet_velocities.output.velocities[-1]
        g.verlet_positions.input.forces = gp.calc_static.output.forces[-1]
        g.verlet_positions.input.masses = ip.structure.get_masses
        g.verlet_positions.input.time_step = ip.time_step
        g.verlet_positions.input.temperature = ip.temperature
        g.verlet_positions.input.temperature_damping_timescale = ip.temperature_damping_timescale

        # reflect individual atoms which stray too far
        g.reflect_atoms.input.default.previous_positions = ip.structure.positions
        g.reflect_atoms.input.default.previous_velocities = gp.initial_velocities.output.velocities[-1]
        g.reflect_atoms.input.default.total_steps = ip.total_steps

        g.reflect_atoms.input.reference_positions = ip.structure.positions
        g.reflect_atoms.input.positions = gp.verlet_positions.output.positions[-1]
        g.reflect_atoms.input.velocities = gp.verlet_positions.output.velocities[-1]
        g.reflect_atoms.input.previous_positions = gp.barostat.output.positions[-1]
        g.reflect_atoms.input.previous_velocities = gp.reflect_atoms.output.velocities[-1]
        g.reflect_atoms.input.structure = ip.structure
        g.reflect_atoms.input.cutoff_distance = gp.cutoff.output.cutoff_distance[-1]
        g.reflect_atoms.input.use_reflection = ip.use_reflection
        g.reflect_atoms.input.total_steps = gp.reflect_atoms.output.total_steps[-1]

        # calc_static
        g.calc_static.input.ref_job_full_path = ip.ref_job_full_path
        g.calc_static.input.structure = gp.barostat.output.structure[-1]
        g.calc_static.input.cell = gp.barostat.output.structure[-1].cell.array
        g.calc_static.input.positions = gp.reflect_atoms.output.positions[-1]

        # verlet_velocities
        g.verlet_velocities.input.velocities = gp.reflect_atoms.output.velocities[-1]
        g.verlet_velocities.input.forces = gp.calc_static.output.forces[-1]
        g.verlet_velocities.input.masses = ip.structure.get_masses
        g.verlet_velocities.input.time_step = ip.time_step
        g.verlet_velocities.input.temperature = ip.temperature
        g.verlet_velocities.input.temperature_damping_timescale = ip.temperature_damping_timescale

        self.set_graph_archive_clock(gp.clock.output.n_counts[-1])


class ProtoConfinedMD(Protocol, ConfinedMD):
    pass


class HarmonicMD(CompoundVertex):
    """
    Runs molecular dynamics, but treats the atoms in the structure as harmonic oscillators. Calculates the forces
        on each atom, and the total potential energy of the structure. If the spring constant is specified, the
        atoms act as Einstein atoms (independent of each other). If the Hessian / force constant matrix is
        specified, the atoms act as Debye atoms.

    Input attributes:
        ref_job_full_path (str): Path to the pyiron job to use for evaluating forces and energies.
        structure (Atoms): The structure evolve.
        temperature (float): Temperature to run at in K.
        n_steps (int): How many MD steps to run for. (Default is 100.)
        temperature_damping_timescale (float): Langevin thermostat timescale in fs. (Default is None, which runs NVE.)
        time_step (float): MD time step in fs. (Default is 1.)
        overheat_fraction (float): The fraction by which to overheat the initial velocities. This can be useful for
            more quickly equilibrating a system whose initial structure is its fully relaxed positions -- in which
            case equipartition of energy tells us that the kinetic energy should be initialized to double the
            desired value. (Default is 2.0, assume energy equipartition is a good idea.)
        spring_constant (float): A single spring / force constant that is used to compute the restoring forces
            on each atom. (Default is 1.)
        force_constants (NxN matrix): The Hessian matrix, obtained from, for ex. Phonopy. (Default is None, treat
            the atoms as independent harmonic oscillators (Einstein atoms.).)

    Output attributes:
        energy_pot (float): Total potential energy of the system in eV.
        energy_kin (float): Total kinetic energy of the system in eV.
        positions (numpy.ndarray): Atomic positions in angstroms.
        velocities (numpy.ndarray): Atomic velocities in angstroms/fs.
        forces (numpy.ndarray): Atomic forces in eV/angstrom. Note: These are the potential gradient forces; thermostat
            forces (if any) are not saved.
    """

    def __init__(self, **kwargs):
        super(HarmonicMD, self).__init__(**kwargs)

        # Protocol defaults
        id_ = self.input.default
        id_.temperature = None
        id_.n_steps = 100
        id_.temperature_damping_timescale = 100.
        id_.time_step = 1.
        id_.overheat_fraction = 2
        id_.spring_constant = None
        id_.force_constants = None

    def define_vertices(self):
        # Graph components
        g = self.graph
        g.initial_velocities = RandomVelocity()
        g.initial_forces = Zeros()
        g.initial_pressures = Zeros()
        g.check_steps = IsGEq()
        g.clock = Counter()
        g.verlet_positions = VerletPositionUpdate()
        g.calc_harmonic = HarmonicHamiltonian()
        g.verlet_velocities = VerletVelocityUpdate()

    def define_execution_flow(self):
        # Execution flow
        g = self.graph
        g.make_pipeline(
            g.initial_velocities,
            g.initial_forces,
            g.initial_pressures,
            g.check_steps, 'false',
            g.verlet_positions,
            g.calc_harmonic,
            g.verlet_velocities,
            g.clock,
            g.check_steps
        )
        g.starting_vertex = g.initial_velocities
        g.restarting_vertex = g.check_steps

    def define_information_flow(self):
        # Data flow
        g = self.graph
        gp = Pointer(self.graph)
        ip = Pointer(self.input)

        # initial_velocities
        g.initial_velocities.input.temperature = ip.temperature
        g.initial_velocities.input.masses = ip.structure.get_masses
        g.initial_velocities.input.overheat_fraction = ip.overheat_fraction

        # initial_forces
        g.initial_forces.input.shape = ip.structure.positions.shape

        # initial_pressures
        g.initial_pressures.input.shape = ip.structure.cell.array.shape

        # check_steps
        g.check_steps.input.target = gp.clock.output.n_counts[-1]
        g.check_steps.input.threshold = ip.n_steps

        # verelt_positions
        g.verlet_positions.input.default.positions = ip.structure.positions
        g.verlet_positions.input.default.velocities = gp.initial_velocities.output.velocities[-1]
        g.verlet_positions.input.default.forces = gp.initial_forces.output.zeros[-1]

        g.verlet_positions.input.positions = gp.verlet_positions.output.positions[-1]
        g.verlet_positions.input.velocities = gp.verlet_velocities.output.velocities[-1]
        g.verlet_positions.input.forces = gp.calc_harmonic.output.forces[-1]
        g.verlet_positions.input.masses = ip.structure.get_masses
        g.verlet_positions.input.time_step = ip.time_step
        g.verlet_positions.input.temperature = ip.temperature
        g.verlet_positions.input.temperature_damping_timescale = ip.temperature_damping_timescale

        # calc_harmonic
        g.calc_harmonic.input.spring_constant = ip.spring_constant
        g.calc_harmonic.input.force_constants = ip.force_constants
        g.calc_harmonic.input.reference_positions = ip.structure.positions
        g.calc_harmonic.input.positions = gp.verlet_positions.output.positions[-1]
        g.calc_harmonic.input.structure = ip.structure

        # verlet_velocities
        g.verlet_velocities.input.velocities = gp.verlet_positions.output.velocities[-1]
        g.verlet_velocities.input.forces = gp.calc_harmonic.output.forces[-1]
        g.verlet_velocities.input.masses = ip.structure.get_masses
        g.verlet_velocities.input.time_step = ip.time_step
        g.verlet_velocities.input.temperature = ip.temperature
        g.verlet_velocities.input.temperature_damping_timescale = ip.temperature_damping_timescale

        self.set_graph_archive_clock(gp.clock.output.n_counts[-1])

    def get_output(self):
        gp = Pointer(self.graph)
        return {
            'energy_pot': ~gp.calc_harmonic.output.energy_pot[-1],
            'energy_kin': ~gp.verlet_velocities.output.energy_kin[-1],
            'positions': ~gp.verlet_positions.output.positions[-1],
            'velocities': ~gp.verlet_velocities.output.velocities[-1],
            'forces': ~gp.calc_harmonic.output.forces[-1],
        }


class ProtoHarmMD(Protocol, HarmonicMD):
    pass



import posixpath
import xml.etree.ElementTree as ET
import re
import itertools

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

from pyiron_base import PyironFactory, DataContainer
from pyiron_contrib.atomistics.atomicrex.function_factory import (
    FunctionFactory,
    FunctionParameterList,
)
from pyiron_contrib.atomistics.atomicrex.utility_functions import write_pretty_xml


class ARPotFactory(PyironFactory):
    """
    Factory class providing convenient acces to fittable potential types.
    TODO: add other potential types supported by atomicrex.
    """

    @staticmethod
    def eam_potential(
        identifier="EAM",
        export_file="output.eam.fs",
        rho_range_factor=2.0,
        resolution=10000,
        species=["*", "*"],
    ):
        return EAMPotential(
            identifier=identifier,
            export_file=export_file,
            rho_range_factor=rho_range_factor,
            resolution=resolution,
            species=species,
        )

    @staticmethod
    def adp_potential(
        identifier="ADP",
        export_file="output.adp.fs",
        rho_range_factor=2.0,
        resolution=10000,
        species=["*", "*"],
    ):
        return ADPotential(
            identifier=identifier,
            export_file=export_file,
            rho_range_factor=rho_range_factor,
            resolution=resolution,
            species=species,
        )

    @staticmethod
    def meam_potential(identifier="MEAM", export_file="meam.out", species=["*", "*"]):
        return MEAMPotential(
            identifier=identifier,
            export_file=export_file,
            species=species,
        )

    @staticmethod
    def lennard_jones_potential(
        sigma, epsilon, cutoff, species={"a": "*", "b": "*"}, identifier="LJ"
    ):
        return LJPotential(
            sigma, epsilon, cutoff, species=species, identifier=identifier
        )

    @staticmethod
    def tersoff_potential(elements, export_file=None):
        return TersoffPotential(elements=elements, export_file=export_file)

    @staticmethod
    def abop_potential(elements, export_file=None):
        return ABOPPotential(elements=elements, export_file=export_file)


class AbstractPotential(DataContainer):
    """Potentials should inherit from this class for hdf5 storage."""

    def __init__(self, init=None, table_name="potential"):
        super().__init__(init, table_name=table_name)

    # def copy_final_to_initial_params(self):
    #    raise NotImplementedError("Should be implemented in the subclass.")

    def write_xml_file(self, directory):
        """
        Write the potential.xml file for the atomicrex job.
        Has to be defined.
        Args:
            directory (str): working directory path
        """
        raise NotImplementedError("Has to be implemented in the subclass")

    def _parse_final_parameters(self, lines):
        raise NotImplementedError("Has to be implemented in the subclass")

    def _potential_as_pd_df(self, job):
        """
        Internal function used to convert a fitted potential in a pandas datafram
        that can be used with lammps calculations.
        Since the formatting is different for every type of potential
        this has to be implemented in the child classes

        Args:
            job (pyiron_job): Takes the fit job as argument, to obtain f.e. the working directory.
        """
        raise NotImplementedError("Should be implemented in the subclass")

    def _plot_final_potential(self):
        raise NotImplementedError("Should be implemented in the subclass")


class BOPAbstract(AbstractPotential):
    def __init__(self, init=None, elements=None, export_file=None, identifier=None):
        super().__init__(init=init)
        if init is None:
            self.identifier = identifier
            if elements is not None:
                if export_file is None:
                    self.export = f"{''.join(elem for elem in elements)}.tersoff"
                else:
                    self.export_file = export_file

                self._tag_dict = _get_tag_dict(elements)
                self.elements = elements
                if self.identifier == "tersoff":
                    self.parameters = TersoffParameters(self._tag_dict)
                elif self.identifier == "abop":
                    self.parameters = ABOPParameters(self._tag_dict)

    @property
    def param_file(self):
        return self.export_file

    def write_xml_file(self, directory):
        self.parameters._write_tersoff_file(directory)

        tersoff = ET.Element(f"{self.identifier}")
        tersoff.set("id", f"{self.identifier}")
        tersoff.set("species-a", "*")
        tersoff.set("species-b", "*")

        output = ET.SubElement(tersoff, "export-potential")
        output.text = f"{self.export_file}"

        params = ET.SubElement(tersoff, "param-file")
        params.text = "input.tersoff"

        tersoff.append(self.parameters._to_xml_element())

        filename = posixpath.join(directory, "potential.xml")
        write_pretty_xml(tersoff, filename)

    def _parse_final_parameters(self, lines):
        for l in lines:
            tag, param, value = _parse_tersoff_line(l)
            self.parameters[tag][param].final_value = value

    def _potential_as_pd_df(self, job):
        """
        Makes the tabulated eam potential written by atomicrex usable
        for pyiron lammps jobs.
        """
        if self.export_file is None:
            raise ValueError("export_file must be set to use the potential with lammps")

        species = [el for el in job.input.atom_types.keys()]
        species_str = ""
        for s in species:
            species_str += f"{s} "

        pot = pd.DataFrame(
            {
                "Name": f"{self.identifier}",
                "Filename": [[f"{job.working_directory}/{self.export_file}"]],
                "Model": ["Custom"],
                "Species": [species],
                "Config": [
                    [
                        "pair_style tersoff\n",
                        f"pair_coeff * * {job.working_directory}/{self.export_file} {species_str}\n",
                    ]
                ],
            }
        )
        return pot


class TersoffPotential(BOPAbstract):
    def __init__(self, init=None, elements=None, export_file=None):
        super().__init__(
            init=init, elements=elements, export_file=export_file, identifier="tersoff"
        )


class ABOPPotential(BOPAbstract):
    def __init__(self, init=None, elements=None, export_file=None):
        super().__init__(
            init=init, elements=elements, export_file=export_file, identifier="abop"
        )


# tag regular expression
tag_re = re.compile("[A-Z][a-z]?")
# Parameters in the order of tersoff file format
tersoff_file_params = [
    "e1",
    "e2",
    "e3",
    "m",
    "gamma",
    "lambda3",
    "c",
    "d",
    "theta0",
    "n",
    "beta",
    "lambda2",
    "B",
    "R",
    "D",
    "lambda1",
    "A",
]

# Transform tersoff to abop parameters
def get_r0(lam1, lam2, A, B):
    return 1 / (lam1 - lam2) * np.log(lam1 * A / (lam2 * B))


def get_D0(lam1, lam2, A, B):
    return (
        A
        * (lam1 / lam2 - 1)
        * np.exp(-lam1 / (lam1 - lam2) * np.log(lam1 * A / lam2 * B))
    )


def get_beta(lam1, lam2):
    return lam1 / np.sqrt(2 * lam1 / lam2)


def get_S(lam1, lam2):
    return lam1 / lam2


# abop to tersoff parameters
def get_lam1(beta, S):
    return beta * np.sqrt(2 * S)


def get_lam2(beta, S):
    return beta * np.sqrt(2 / S)


def get_A(beta, S, D0, r0):
    return D0 / (S - 1) * np.exp(beta * np.sqrt(2 * S) * r0)


def get_B(beta, S, D0, r0):
    return D0 * S / (S - 1) * np.exp(beta * np.sqrt(2 / S) * r0)


class BOPParameters(DataContainer):
    def __init__(self, tag_dict=None):
        super().__init__()
        if tag_dict is not None:
            self._setup_parameters(tag_dict)

    def _setup_parameters(self, tag_dict):
        raise NotImplementedError("Implement in subclass")

    @staticmethod
    def tags_from_elements(elements):
        """
        Helper function that returns a list of tags
        from a list of elements, f.e. for the start_vals_from_file function

        Args:
            elements (list[str]): list of elements .f.e ["Si", "C"]

        Returns:
            list[str]: list of tags f.e. ["SiSiSi", "SiSiC", ...]
        """
        return _tag_list(elements)


class TersoffParameters(BOPParameters):
    def __init__(self, tag_dict=None):
        super().__init__(tag_dict)

    def _setup_parameters(self, tag_dict):
        for tag, v in tag_dict.items():
            self[tag] = FunctionParameterList()

            # Parameters are added in the same order as they appear in lammps/tersoff format
            # If this order is kept always? this allows to iterate through it in a for loop
            # in some sitauations, instead of looking up the keys
            # 3 body params m(not fittable), gamma, lambda3, c, d, theta0
            self[tag].add_parameter("m", start_val=3, fitable=False)  # 3 or 1

            self[tag].add_parameter("gamma", start_val=None, tag=tag)
            self[tag].add_parameter("lambda3", start_val=None, tag=tag)
            self[tag].add_parameter("c", start_val=None, tag=tag)
            self[tag].add_parameter("d", start_val=None, tag=tag)
            self[tag].add_parameter("theta0", start_val=None, tag=tag)

            # 2 body params powern(only abop fittable), beta, lambda1, lambda2, A, B
            self[tag].add_parameter("n", start_val=None, fitable=False)
            self[tag].add_parameter("beta", start_val=None, fitable=v, tag=tag)
            self[tag].add_parameter("lambda2", start_val=None, fitable=v, tag=tag)
            self[tag].add_parameter("B", start_val=None, fitable=v, tag=tag)
            # 2 and 3 body params D and R
            self[tag].add_parameter("D", start_val=None, fitable=False)
            self[tag].add_parameter("R", start_val=None, fitable=False)
            # 2 body params again
            self[tag].add_parameter("lambda1", start_val=None, fitable=v, tag=tag)
            self[tag].add_parameter("A", start_val=None, fitable=v, tag=tag)

    def start_vals_from_file(self, file, tag_list=None):
        with open(file) as f:
            for l in f:
                l = l.strip()
                if not l:
                    continue
                elif l.startswith("#"):
                    continue
                else:
                    l = l.split()
                    tag = "".join((l[0], l[1], l[2]))
                    ## Allow to filter the file for certain tags by setting a tag_list
                    if tag_list is not None:
                        if tag not in tag_list:
                            continue

                    for i in range(3, 17):
                        param = tersoff_file_params[i]
                        try:
                            self[tag][param].start_val = float(l[i])
                        except:
                            KeyError(
                                f"Could not assign start value of tag: {tag} to parameter: {param}"
                            )

    def _write_tersoff_file(self, directory, filename="input.tersoff"):
        filepath = posixpath.join(directory, filename)
        with open(filepath, "w") as f:
            f.write(
                "# Tersoff parameter file in lammps/tersoff format written via pyiron atomicrex interface.\n"
            )
            f.write("# Necessary to fit tersoff and abop potentials.\n#\n")
            f.write(f"# {' '.join(tersoff_file_params)}\n#\n")
            for tag in self.keys():
                for el in tag_re.findall(tag):
                    f.write(f"{el} ")
                for i in range(3, 17):
                    param = tersoff_file_params[i]
                    f.write(f"{self[tag][param].start_val} ")
                f.write("\n")

    def _to_xml_element(self):
        fit_dof = ET.Element("fit-dof")
        for tag, parameters in self.items():
            for param in parameters.values():
                if param.fitable:
                    fit_dof.append(param._to_xml_element())
        return fit_dof


class ABOPParameters(BOPParameters):
    def __init__(self, tag_dict=None):
        super().__init__(tag_dict)

    def _setup_parameters(self, tag_dict):
        for tag, v in tag_dict.items():
            self[tag] = FunctionParameterList()

            # 2 Body parameters, fitable only if v (see https://lammps.sandia.gov/doc/pair_tersoff.html)
            self[tag].add_parameter("D0", start_val=None, fitable=v, tag=tag)
            self[tag].add_parameter("S", start_val=None, fitable=v, tag=tag)
            self[tag].add_parameter("r0", start_val=None, fitable=v, tag=tag)
            self[tag].add_parameter("beta", start_val=None, fitable=v, tag=tag)
            self[tag].add_parameter("beta2", start_val=None, fitable=v, tag=tag)
            # default enable False for n in 1/2n term and standard value 1
            self[tag].add_parameter(
                "powern", start_val=1, fitable=v, enabled=False, tag=tag
            )
            # 3 Body parameters
            self[tag].add_parameter("gamma", start_val=None, tag=tag)
            self[tag].add_parameter("c", start_val=None, tag=tag)
            self[tag].add_parameter("d", start_val=None, tag=tag)
            self[tag].add_parameter("h", start_val=None, tag=tag)
            # twomu = lambda3?
            self[tag].add_parameter("twomu", start_val=None, tag=tag)
            # m is not fittable
            self[tag].add_parameter(
                "m", start_val=1.0, fitable=False, tag=tag
            )  # 3 or 1
            # 2 and 3 body
            # default enable False for cutoff parameters
            self[tag].add_parameter("bigd", start_val=None, enabled=False, tag=tag)
            self[tag].add_parameter("bigr", start_val=None, enabled=False, tag=tag)

    def start_vals_from_file(self, file, tag_list=None):
        if tag_list is None:
            tag_list = self.list_groups()
        tag_dict = _tagdict_from_taglist(tag_list)
        t_params = TersoffParameters(tag_dict)
        t_params.start_vals_from_file(file, tag_list=tag_list)
        for tag, params in t_params.items():
            lam1 = params["lambda1"].start_val
            lam2 = params["lambda2"].start_val
            A = params["A"].start_val
            B = params["B"].start_val

            self[tag]["h"].start_val = -params["theta0"].start_val

            if tag_dict[tag]:
                self[tag]["r0"].start_val = get_r0(lam1, lam2, A, B)
                self[tag]["D0"].start_val = get_D0(lam1, lam2, A, B)
                self[tag]["S"].start_val = get_S(lam1, lam2)
                self[tag]["beta"].start_val = get_beta(lam1, lam2)
            else:  # pair only parameters are not read for these tags and are set to 0 to prevent 0 division errors
                self[tag]["r0"].start_val = 0.0
                self[tag]["D0"].start_val = 0.0
                self[tag]["S"].start_val = 0.0
                self[tag]["beta"].start_val = 0.0

            # Other parameters
            self[tag]["m"].start_val = params["m"].start_val
            self[tag]["powern"].start_val = params["n"].start_val
            self[tag]["c"].start_val = params["c"].start_val
            self[tag]["d"].start_val = params["d"].start_val
            self[tag]["bigd"].start_val = params["D"].start_val
            self[tag]["bigr"].start_val = params["R"].start_val
            self[tag]["gamma"].start_val = params["gamma"].start_val
            self[tag]["twomu"].start_val = params["lambda3"].start_val
            self[tag]["beta2"].start_val = params["beta"].start_val

    def _write_tersoff_file(self, directory, filename="input.tersoff"):
        tersoff = abop_to_tersoff_params(self)
        tersoff._write_tersoff_file(directory, filename=filename)

    def _to_xml_element(self):
        fit_dof = ET.Element("fit-dof")
        for tag, parameters in self.items():
            for param in parameters.values():
                if param.fitable:
                    fit_dof.append(param._to_xml_element())
        return fit_dof


def abop_to_tersoff_params(abop_parameters):
    """
    Transforms a ABOPParameters instance to TersoffParameters
    Does not take care of min or max vals, only initial values!
    Necessary to write the tersoff file as input.
    Not tested for anything else.
    Args:
        tersoff_parameters ([type]): [description]
    """

    tag_dict = _tagdict_from_taglist(abop_parameters.groups())
    tersoff = TersoffParameters(tag_dict)

    for tag, params in tersoff.items():
        if tag_dict[tag]:
            beta = abop_parameters[tag]["beta"].start_val
            S = abop_parameters[tag]["S"].start_val
            D0 = abop_parameters[tag]["D0"].start_val
            r0 = abop_parameters[tag]["r0"].start_val

            # parameters that need calculations
            params["A"].start_val = get_A(beta, S, D0, r0)
            params["B"].start_val = get_B(beta, S, D0, r0)
            params["lambda1"].start_val = get_lam1(beta, S)
            params["lambda2"].start_val = get_lam2(beta, S)
        else:  # pair only parameters are not read for these tags and are set to 0 to prevent 0 division errors
            params["A"].start_val = 0.0
            params["B"].start_val = 0.0
            params["lambda1"].start_val = 0.0
            params["lambda2"].start_val = 0.0

        params["theta0"].start_val = -abop_parameters[tag]["h"].start_val

        # Other parameters
        params["m"].start_val = abop_parameters[tag]["m"].start_val
        params["n"].start_val = abop_parameters[tag]["powern"].start_val
        params["c"].start_val = abop_parameters[tag]["c"].start_val
        params["d"].start_val = abop_parameters[tag]["d"].start_val
        params["D"].start_val = abop_parameters[tag]["bigd"].start_val
        params["R"].start_val = abop_parameters[tag]["bigr"].start_val
        params["gamma"].start_val = abop_parameters[tag]["gamma"].start_val
        params["lambda3"].start_val = abop_parameters[tag]["twomu"].start_val
        params["beta"].start_val = abop_parameters[tag]["beta2"].start_val
    return tersoff


class LJPotential(AbstractPotential):
    """
    Lennard Jones potential. Lacking some functionality,
    mostly implemented for testing and demonstatration purposes.
    TODO: Check if this still works after alls changes
    TODO: Implement missing functionality.
    """

    def __init__(
        self,
        init=None,
        sigma=None,
        epsilon=None,
        cutoff=None,
        species=None,
        identifier=None,
    ):
        super().__init__(init=init)
        if init is None:
            self.sigma = sigma
            self.epsilon = epsilon
            self.cutoff = cutoff
            self.species = species
            self.identifier = identifier

    def write_xml_file(self, directory):
        """
        Internal function to create xml element.
        """
        lj = ET.Element("lennard-jones")
        lj.set("id", f"{self.identifier}")
        lj.set("species-a", self.species["a"])
        lj.set("species-b", self.species["b"])

        sigma = ET.SubElement(lj, "sigma")
        sigma.text = f"{self.sigma}"
        eps = ET.SubElement(lj, "epsilon")
        eps.text = f"{self.epsilon}"
        cutoff = ET.SubElement(lj, "cutoff")
        cutoff.text = f"{self.cutoff}"

        fit_dof = ET.SubElement(lj, "fit-dof")
        sigma = ET.SubElement(fit_dof, "sigma")
        epsilon = ET.SubElement(fit_dof, "epsilon")

        filename = posixpath.join(directory, "potential.xml")
        write_pretty_xml(lj, filename)


class EAMlikeMixin:
    def copy_final_to_initial_params(self, filter_func=None):
        """
        Copies final values of function paramters to start values.
        This f.e. allows to continue global with local minimization.
        """
        for functions in self._function_tuple:
            for f in functions.values():
                f.copy_final_to_initial_params(filter_func=filter_func)

    def lock_parameters(self, filter_func=None):
        for functions in self._function_tuple:
            for f in functions.values():
                f.lock_parameters(filter_func=filter_func)

    def count_parameters(self, enabled_only=True):
        parameters = 0
        for functions in self._function_tuple:
            for f in functions.values():
                parameters += f.count_parameters(enabled_only=enabled_only)
        return parameters

    @property
    def _function_tuple(self):
        raise NotImplementedError("Implement a tuple with functions in subclass")

    @property
    def _function_dict(self):
        raise NotImplementedError("Implement a tuple with functions in subclass")

    def _mapping_functions_xml(self, pot):
        mappingxml = ET.SubElement(pot, "mapping")
        functionsxml = ET.SubElement(pot, "functions")

        for k, functions in self._function_dict.items():
            for f in functions.values():
                fxml = ET.SubElement(mappingxml, k)
                if len(f.species) == 1:
                    fxml.set("species", f"{f.species[0]}")
                elif len(f.species) == 2:
                    fxml.set("species-a", f"{f.species[0]}")
                    fxml.set("species-b", f"{f.species[1]}")
                elif len(f.species) == 2:
                    fxml.set("species-a", f"{f.species[0]}")
                    fxml.set("species-b", f"{f.species[1]}")
                    fxml.set("species-c", f"{f.species[2]}")
                fxml.set("function", f"{f.identifier}")
                functionsxml.append(f._to_xml_element())

    def _eam_parse_final_parameters(self, lines):
        """
        Internal Function.
        Parse function parameters from atomicrex output.

        Args:
            lines (list[str]): atomicrex output lines

        Raises:
            KeyError: Raises if a parsed parameter can't be matched to a function.
        """
        for l in lines:
            identifier, leftover, value = _parse_parameter_line(l)
            found = False
            for functions in self._function_tuple:
                if identifier in functions:
                    functions[identifier]._parse_final_parameter(leftover, value)
                    found = True
            if not found:
                raise KeyError(
                    f"Can't find {identifier} in potential, probably something went wrong during parsing.\n"
                    "Fitting parameters of screening functions probably doesn't work right now"
                )


class EAMPotential(AbstractPotential, EAMlikeMixin):
    """
    Embedded-Atom-Method potential.
    Usage: Create using the potential factory class.
    Add functions defined using the function_factory
    to self.pair_interactions, self.electron_densities
    and self.embedding_energies in dictionary style,
    using the identifier of the function as key.
    Example:
    eam.pair_interactions["V"] = morse_function

    """

    def __init__(
        self,
        init=None,
        identifier=None,
        export_file=None,
        rho_range_factor=None,
        resolution=None,
        species=None,
    ):

        super().__init__(init=init)
        if init is None:
            self.pair_interactions = DataContainer(table_name="pair_interactions")
            self.electron_densities = DataContainer(table_name="electron_densities")
            self.embedding_energies = DataContainer(table_name="embedding_energies")
            self.identifier = identifier
            self.export_file = export_file
            self.rho_range_factor = rho_range_factor
            self.resolution = resolution
            self.species = species

    @property
    def _function_tuple(self):
        return (
            self.pair_interactions,
            self.electron_densities,
            self.embedding_energies,
        )

    def _potential_as_pd_df(self, job):
        """
        Makes the tabulated eam potential written by atomicrex usable
        for pyiron lammps jobs.
        """
        if self.export_file is None:
            raise ValueError("export_file must be set to use the potential with lammps")

        species = [el for el in job.input.atom_types.keys()]
        species_str = ""
        for s in species:
            species_str += f"{s} "

        pot = pd.DataFrame(
            {
                "Name": f"{self.identifier}",
                "Filename": [[f"{job.working_directory}/{self.export_file}"]],
                "Model": ["Custom"],
                "Species": [species],
                "Config": [
                    [
                        "pair_style eam/fs\n",
                        f"pair_coeff * * {job.working_directory}/{self.export_file} {species_str}\n",
                    ]
                ],
            }
        )
        return pot

    def write_xml_file(self, directory):
        """
        Internal function to convert to an xml element
        """
        eam = ET.Element("eam")
        eam.set("id", f"{self.identifier}")
        eam.set("species-a", f"{self.species[0]}")
        eam.set("species-b", f"{self.species[1]}")

        if self.export_file:
            export = ET.SubElement(eam, "export-eam-file")
            export.set("resolution", f"{self.resolution}")
            export.set("rho-range-factor", f"{self.rho_range_factor}")
            export.text = f"{self.export_file}"

        mapping = ET.SubElement(eam, "mapping")
        functions = ET.SubElement(eam, "functions")

        for pot in self.pair_interactions.values():
            pair_interaction = ET.SubElement(mapping, "pair-interaction")
            pair_interaction.set("species-a", f"{pot.species[0]}")
            pair_interaction.set("species-b", f"{pot.species[1]}")
            pair_interaction.set("function", f"{pot.identifier}")
            functions.append(pot._to_xml_element())

        for pot in self.electron_densities.values():
            electron_density = ET.SubElement(mapping, "electron-density")
            electron_density.set("species-a", f"{pot.species[0]}")
            electron_density.set("species-b", f"{pot.species[1]}")
            electron_density.set("function", f"{pot.identifier}")
            functions.append(pot._to_xml_element())

        for pot in self.embedding_energies.values():
            embedding_energy = ET.SubElement(mapping, "embedding-energy")
            embedding_energy.set("species", f"{pot.species[0]}")
            embedding_energy.set("function", f"{pot.identifier}")
            functions.append(pot._to_xml_element())

        filename = posixpath.join(directory, "potential.xml")
        write_pretty_xml(eam, filename)

    def _parse_final_parameters(self, lines):
        """
        Internal Function.
        Parse function parameters from atomicrex output.

        Args:
            lines (list[str]): atomicrex output lines

        Raises:
            KeyError: Raises if a parsed parameter can't be matched to a function.
        """
        for l in lines:
            identifier, leftover, value = _parse_parameter_line(l)
            if identifier in self.pair_interactions:
                self.pair_interactions[identifier]._parse_final_parameter(
                    leftover, value
                )
            elif identifier in self.electron_densities:
                self.electron_densities[identifier]._parse_final_parameter(
                    leftover, value
                )
            elif identifier in self.embedding_energies:
                self.embedding_energies[identifier]._parse_final_parameter(
                    leftover, value
                )
            else:

                raise KeyError(
                    f"Can't find {identifier} in potential, probably something went wrong during parsing.\n"
                    "Fitting parameters of screening functions probably doesn't work right now"
                )

    def plot_final_potential(self, job, filename=None):
        """
        Plots the the fitted potential. Reads the necessary data from eam.fs file (funcfl format used by lammps),
        therefore does not work if no table is written. Can be used with disabled fitting to plot functions that
        can't be plotted directly like splines.

        Args:
            job (AtomicrexJob): Instance of the job to construct filename.
            filename (str, optional): If a filename is given this eam is plotted instead of the fitted one. Defaults to None.
        """
        if filename is None:
            filename = f"{job.working_directory}/{self.export_file}"

        elements, r_values, rho_values = read_eam_fs_file(filename=filename)

        # TODO: figure out how to index ax for multiple elements
        fig, ax = plt.subplots(
            nrows=3 * len(elements),
            ncols=len(elements),
            figsize=(len(elements) * 8, len(elements) * 3 * 6),
            squeeze=False,
        )
        for i, (el, pot_dict) in enumerate(elements.items()):
            V_count = 0
            rho_count = 0
            for pot, y in pot_dict.items():
                if pot == "F":
                    xdata = rho_values
                    xlabel = "$\\rho $ [a.u.]"
                    k = 0
                    ylim = (np.min(y) - 0.5, 5)
                    ax[i * 3 + k, 0].plot(xdata, y)
                    ax[i * 3 + k, 0].set(ylim=ylim, title=f"{el} {pot}", xlabel=xlabel)
                elif "rho" in pot:
                    xdata = r_values
                    xlabel = "r [$\AA$]"
                    k = 1
                    ylim = (np.min(y) - 0.1, 1)
                    ax[i * 3 + k, rho_count].plot(xdata, y)
                    ax[i * 3 + k, rho_count].set(
                        ylim=ylim, title=f"{el} {pot}", xlabel=xlabel
                    )
                    rho_count += 1
                elif "V" in pot:
                    xdata = r_values[1:]
                    y = y[1:]
                    xlabel = "r [$\AA$]"
                    k = 2
                    ylim = (np.min(y) - 0.1, 2)
                    ax[i * 3 + k, V_count].plot(xdata, y)
                    ax[i * 3 + k, V_count].set(
                        ylim=ylim, title=f"{el} {pot}", xlabel=xlabel
                    )
                    V_count += 1
        return fig, ax

    def count_local_extrema(
        self, job, filename=None, count_minima=True, count_maxima=False
    ):
        if filename is None:
            filename = f"{job.working_directory}/{self.export_file}"
        elements, _r_values, _rho_values = read_eam_fs_file(filename=filename)
        extrema_dict = {}
        for el, func in elements.items():
            extrema_dict[el] = {}
            for func_name, a in func:
                extrema = 0
                if count_minima:
                    min_arr = np.r_[True, a[1:] < a[:-1]] & np.r_[a[:-1] < a[1:], True]
                    extrema += len(min_arr[min_arr])
                if count_maxima:
                    max_arr = np.r_[True, a[1:] > a[:-1]] & np.r_[a[:-1] > a[1:], True]
                    extrema += len(max_arr[max_arr])
                extrema_dict[el][func_name] = extrema
        return extrema_dict


class ADPotential(AbstractPotential, EAMlikeMixin):
    """
    Angular dependent potential.
    Usage: Create using the potential factory class.
    Add functions defined using the function_factory
    to self.pair_interactions, self.electron_densities
    and self.embedding_energies, self.u_functions and
    self.w_functions in dictionary style,
    using the identifier of the function as key.
    Example:
    eam.pair_interactions["V"] = morse_function

    """

    def __init__(
        self,
        init=None,
        identifier=None,
        export_file=None,
        rho_range_factor=None,
        resolution=None,
        species=None,
    ):

        super().__init__(init=init)
        if init is None:
            self.pair_interactions = DataContainer(table_name="pair_interactions")
            self.electron_densities = DataContainer(table_name="electron_densities")
            self.embedding_energies = DataContainer(table_name="embedding_energies")
            self.u_functions = DataContainer(table_name="u_functions")
            self.w_functions = DataContainer(table_name="w_functions")
            self.identifier = identifier
            self.export_file = export_file
            self.rho_range_factor = rho_range_factor
            self.resolution = resolution
            self.species = species

    @property
    def _function_tuple(self):
        return (
            self.pair_interactions,
            self.electron_densities,
            self.embedding_energies,
            self.u_functions,
            self.w_functions,
        )

    @property
    def _function_dict(self):
        return {
            "pair-interaction": self.pair_interactions,
            "electron-density": self.electron_densities,
            "embedding-energy": self.embedding_energies,
            "u-function": self.u_functions,
            "w-function": self.w_functions,
        }

    def _potential_as_pd_df(self, job):
        """
        Makes the tabulated eam potential written by atomicrex usable
        for pyiron lammps jobs.
        """
        if self.export_file is None:
            raise ValueError("export_file must be set to use the potential with lammps")

        species = [el for el in job.input.atom_types.keys()]
        species_str = ""
        for s in species:
            species_str += f"{s} "

        pot = pd.DataFrame(
            {
                "Name": f"{self.identifier}",
                "Filename": [[f"{job.working_directory}/{self.export_file}"]],
                "Model": ["Custom"],
                "Species": [species],
                "Config": [
                    [
                        "pair_style adp\n",
                        f"pair_coeff * * {job.working_directory}/{self.export_file} {species_str}\n",
                    ]
                ],
            }
        )
        return pot

    def write_xml_file(self, directory):
        """
        Internal function to convert to an xml element
        """
        adp = ET.Element("adp")
        adp.set("id", f"{self.identifier}")
        adp.set("species-a", f"{self.species[0]}")
        adp.set("species-b", f"{self.species[1]}")

        if self.export_file:
            export = ET.SubElement(adp, "export-adp-file")
            export.set("resolution", f"{self.resolution}")
            export.set("rho-range-factor", f"{self.rho_range_factor}")
            export.text = f"{self.export_file}"

        self._mapping_functions_xml(adp)

        filename = posixpath.join(directory, "potential.xml")
        write_pretty_xml(adp, filename)

    def _parse_final_parameters(self, lines):
        return self._eam_parse_final_parameters(lines)


class MEAMPotential(AbstractPotential, EAMlikeMixin):
    def __init__(self, init=None, identifier=None, export_file=None, species=None):
        super().__init__(init=init)
        if init is None:
            self.pair_interactions = DataContainer(table_name="pair_interactions")
            self.electron_densities = DataContainer(table_name="electron_densities")
            self.embedding_energies = DataContainer(table_name="embedding_energies")
            self.f_functions = DataContainer(table_name="f_functions")
            self.g_functions = DataContainer(table_name="g_functions")
            self.identifier = identifier
            self.export_file = export_file
            self.species = species

    @property
    def _function_tuple(self):
        return (
            self.pair_interactions,
            self.electron_densities,
            self.embedding_energies,
            self.f_functions,
            self.g_functions,
        )

    def write_xml_file(self, directory):
        """
        Internal function to convert to an xml element
        """
        meam = ET.Element("meam")
        meam.set("id", f"{self.identifier}")
        meam.set("species-a", f"{self.species[0]}")
        meam.set("species-b", f"{self.species[1]}")

        if self.export_file:
            export = ET.SubElement(meam, "export-functions")
            export.text = f"{self.export_file}"

        mapping = ET.SubElement(meam, "mapping")
        functions = ET.SubElement(meam, "functions")

        for pot in self.pair_interactions.values():
            pair_interaction = ET.SubElement(mapping, "pair-interaction")
            pair_interaction.set("species-a", f"{pot.species[0]}")
            pair_interaction.set("species-b", f"{pot.species[1]}")
            pair_interaction.set("function", f"{pot.identifier}")
            functions.append(pot._to_xml_element())

        for pot in self.electron_densities.values():
            electron_density = ET.SubElement(mapping, "electron-density")
            electron_density.set("species-a", f"{pot.species[0]}")
            electron_density.set("species-b", f"{pot.species[1]}")
            electron_density.set("function", f"{pot.identifier}")
            functions.append(pot._to_xml_element())

        for pot in self.f_functions.values():
            f_function = ET.SubElement(mapping, "f-function")
            f_function.set("species-a", f"{pot.species[0]}")
            f_function.set("species-b", f"{pot.species[1]}")
            f_function.set("function", f"{pot.identifier}")
            functions.append(pot._to_xml_element())

        for pot in self.g_functions.values():
            g_function = ET.SubElement(mapping, "g-function")
            g_function.set("species-a", f"{pot.species[0]}")
            g_function.set("species-b", f"{pot.species[1]}")
            g_function.set("species-c", f"{pot.species[2]}")
            g_function.set("function", f"{pot.identifier}")
            functions.append(pot._to_xml_element())

        for pot in self.embedding_energies.values():
            embedding_energy = ET.SubElement(mapping, "embedding-energy")
            embedding_energy.set("species", f"{pot.species[0]}")
            embedding_energy.set("function", f"{pot.identifier}")
            functions.append(pot._to_xml_element())

        filename = posixpath.join(directory, "potential.xml")
        write_pretty_xml(meam, filename)

    def _parse_final_parameters(self, lines):
        """
        Internal Function.
        Parse function parameters from atomicrex output.

        Args:
            lines (list[str]): atomicrex output lines

        Raises:
            KeyError: Raises if a parsed parameter can't be matched to a function.
        """
        for l in lines:
            identifier, leftover, value = _parse_parameter_line(l)
            if identifier in self.pair_interactions:
                self.pair_interactions[identifier]._parse_final_parameter(
                    leftover, value
                )
            elif identifier in self.electron_densities:
                self.electron_densities[identifier]._parse_final_parameter(
                    leftover, value
                )
            elif identifier in self.f_functions:
                self.f_functions[identifier]._parse_final_parameter(leftover, value)
            elif identifier in self.g_functions:
                self.g_functions[identifier]._parse_final_parameter(leftover, value)
            elif identifier in self.embedding_energies:
                self.embedding_energies[identifier]._parse_final_parameter(
                    leftover, value
                )
            else:
                raise KeyError(
                    f"Can't find {identifier} in potential, probably something went wrong during parsing.\n"
                    "Fitting parameters of screening functions probably doesn't work right now"
                )


def _parse_parameter_line(line):
    """
    Internal Function.
    Parses the function identifier, leftover and final value of a function parameter
    from an atomicrex output line looking like:
    EAM[EAM].V_CuCu[morse-B].delta: 0.0306585 [-0.5:0.5]
    EAM[EAM].CuCu_rho[spline].node[1].y: 2.10988 [1:3]
    EAM[EAM].V_CuCu[functionSum].V_rep[user-function].A: 0.680291
    EAM[EAM].CuCu_rho[spline].derivative-left: -1000.19
    Returns tuples like these:
    ("V_CuCu", ["delta:"] ,0.0306585)
    ("CuCu_rho", ["node[1", "y:"] ,2.10988)
    ("V_CuCu", ["V_rep[user-function", "A:"],0.680291)
    ("CuCu_rho", ["derivative-left:"] ,-1000.19)
    This allows to flexibly parse the lines for the different types
    of functional forms by passing leftover to a function defined
    in the corresponding function class.

    TODO: Add parsing of polynomial parameters.
    Returns:
        (str, [str], float): identifier, leftover, value
    """
    line = line.strip().split()
    value = float(line[1])
    info = line[0].split("].")
    identifier = info[1].split("[")[0]
    leftover = info[2:]
    return identifier, leftover, value


def _tag_list(elements):
    """
    Internal helper function that returns a list of tags
    from a list of elements

    Args:
        elements (list[str]): list of elements .f.e ["Si", "C"]

    Returns:
        list[str]: list of tags f.e. ["SiSiSi", "SiSiC", ...]
    """
    return ["".join(el) for el in itertools.product(elements, repeat=3)]


def _get_tag_dict(elements):
    """
    Internal helpfer function for Tersoff and ABOP potentials.
    Uses a list of elements strings to return a dict of all tags and if these tags need 2 and 3 body parameters (True)
    or only 3 body parameters (False).
    F.e. ["Si", "C"] returns:
        {'SiSiSi': True,
        'SiSiC': False,
        'SiCSi': False,
        'SiCC': True,
        'CSiSi': True,
        'CSiC': False,
        'CCSi': False,
        'CCC': True}

    Args:
        list[str]: list of elements

    Returns:
        dict[dict]: elemen tags and
    """
    tags = {}
    tag_list = list(itertools.product(elements, repeat=3))
    for l in tag_list:
        k = "".join(l)
        if l[1] == l[2]:
            tags[k] = True  # True if 2 and 3 body
        else:
            tags[k] = False  # False if only 3 body
    return tags


def _parse_tersoff_line(line):
    """
    Internal Function.
    Parses the tag, parameter and final value of a function parameter
    from an atomicrex output line looking like:

    tersoff[Tersoff].lambda1[SiSiSi]: 2.4799 [0:]

    Returns:
        [(str, str, float): tag, param, value
    """
    line = line.strip().split()
    value = float(line[1])
    info = line[0].split("].")[1]
    info = info.split("[")
    param = info[0]
    tag = info[1].rstrip("]:")
    return tag, param, value


def _tagdict_from_taglist(taglist):
    tagdict = {}
    for tag in taglist:
        elems = tag_re.findall(tag)
        if elems[1] == elems[2]:
            tagdict[tag] = True
        else:
            tagdict[tag] = False
    return tagdict


def read_eam_fs_file(filename):
    # Read the cotents of an eam/fs file and returns it as a dicitionary of numpy arrays
    # this is copy pasted from one of my old scripts and could probably be reworked
    elements = {}
    with open(filename, "r") as f:
        # skip comment lines
        for _ in range(3):
            f.readline()

        element_list = f.readline().split()
        for element in element_list[1:]:
            elements[element] = {}

        Nrho, drho, Nr, dr, cutoff = f.readline().split()
        Nrho = int(Nrho)
        drho = float(drho)
        Nr = int(Nr)
        dr = float(dr)
        cutoff = float(cutoff)

        rho_values = np.linspace(0, Nrho * drho, Nrho, endpoint=False)
        r_values = np.linspace(0, Nr * dr, Nr, endpoint=False)

        for element in elements:
            # skip a line with unnecessary information
            f.readline()
            elements[element]["F"] = np.fromfile(f, count=Nrho, sep=" ")
            for rho_element in elements:
                elements[element][
                    "rho_{}{}".format(element, rho_element)
                ] = np.fromfile(f, count=Nr, sep=" ")

        # V_ij = V_ji so it is written only once in the file => avoid attempts to read it twice
        # with a list of elements where it has been read
        # TODO: Have another look how to do this checking
        V_written = []
        for element in elements:
            for V_element in elements:
                elementV_element = "{}{}".format(element, V_element)
                V_elementelement = "{}{}".format(V_element, element)
                if (
                    not elementV_element in V_written
                    and not V_elementelement in V_written
                ):
                    elements[element]["V_{}".format(elementV_element)] = np.fromfile(
                        f, count=Nr, sep=" "
                    )
                    # The tabulated values are not V(r) but V(r) * r, so they are divided by r here,
                    # with exception of the first value to prevent division by 0.
                    elements[element]["V_{}".format(elementV_element)][1:] = (
                        elements[element]["V_{}".format(elementV_element)][1:]
                        / r_values[1:]
                    )
                    V_written.append(elementV_element)
    return elements, r_values, rho_values



import xml.etree.ElementTree as ET
import posixpath

import numpy as np
from ase.data import atomic_masses, atomic_numbers

from pyiron_base import DataContainer, PyironFactory

from pyiron_contrib.atomistics.atomicrex.utility_functions import write_pretty_xml
from pyiron_contrib.atomistics.atomicrex.potential_factory import ARPotFactory
from pyiron_contrib.atomistics.atomicrex.function_factory import FunctionFactory
from pyiron_contrib.atomistics.atomicrex.parameter_constraints import (
    ParameterConstraints,
)


class GeneralARInput(DataContainer):
    """
    Class to store general input of an atomicrex job,
    f.e. the fit algorithm.
    """

    __version__ = "0.1.0"

    def __init__(
        self,
        table_name="general_input",
        name="Atomicrex Job",
        verbosity="medium",
        real_precision=16,
        validate_potentials=False,
        atom_types=None,
        fit_algorithm=None,
        output_interval=10000,
        enable_fitting=True,
    ):

        super().__init__(
            table_name="general_input",
        )

        self.name = name
        self.verbosity = verbosity
        self.real_precision = real_precision
        self.validate_potentials = validate_potentials
        self.atom_types = AtomTypes()
        self.fit_algorithm = fit_algorithm
        self.output_interval = output_interval
        self.enable_fitting = enable_fitting
        # version "0.1.0"
        self.output_file = "atomicrex.out"
        self.parameter_constraints = ParameterConstraints()

    def _write_xml_file(self, directory, job=None):
        """Internal function.
        Write the main input xml file in a directory.

        Args:
            directory (str): Working directory
        """
        root = ET.Element("job")

        output_file = ET.SubElement(root, "output-file")
        output_file.text = self.output_file

        name = ET.SubElement(root, "name")
        name.text = self.name

        verbosity = ET.SubElement(root, "verbosity")
        verbosity.text = self.verbosity

        if self.validate_potentials:
            validate_potentials = ET.SubElement(root, "validate-potentials")

        real_precision = ET.SubElement(root, "real-precision")
        real_precision.text = f"{self.real_precision}"

        atom_types = ET.SubElement(root, "atom-types")
        if len(self.atom_types) == 0:
            eles = job.structures._structures.get_elements()
            for ele in eles:
                self.atom_types[ele] = None

        for k, v in self.atom_types.items():
            species = ET.SubElement(atom_types, "species")
            species.text = k
            if v is not None:
                mass, index = v
            else:
                index = atomic_numbers[k]
                mass = atomic_masses[index]

            species.set("mass", f"{mass}")
            species.set("atomic-number", f"{index}")

        if not isinstance(self.fit_algorithm, ScipyAlgorithm):
            fitting = ET.SubElement(root, "fitting")
            if self.enable_fitting:
                fitting.set("enabled", "true")
            else:
                fitting.set("enabled", "false")
            fitting.set("output-interval", f"{self.output_interval}")
            fitting.append(self.fit_algorithm._to_xml_element())

        potentials = ET.SubElement(root, "potentials")
        include = ET.SubElement(potentials, "xi:include")
        include.set("href", "potential.xml")
        include.set("xmlns:xi", "http://www.w3.org/2003/XInclude")

        structures = ET.SubElement(root, "structures")
        include = ET.SubElement(structures, "xi:include")
        include.set("href", "structures.xml")
        include.set("xmlns:xi", "http://www.w3.org/2003/XInclude")

        if len(self.parameter_constraints) > 0:
            root.append(self.parameter_constraints._to_xml_element())

        file_name = posixpath.join(directory, "main.xml")
        write_pretty_xml(root, file_name)


class AtomTypes(DataContainer):
    """
    DataContainer used to specify elements in the fit job.
    Elements can be added using dictionary or attribute syntax.
    Their value should None or a (mass, index) tuple as value.
    If value is None the mass and index are taken from the ase package.
    Examples:
    job.input.atom_types.Cu = None
    job.input.atom_types["Cu"] = None
    job.input.atom_types.Cu = (63.546, 29)
    """

    def __init__(self, *args, **kwargs):
        super().__init__(table_name="AtomTypes", *args, **kwargs)


class AlgorithmFactory(PyironFactory):
    """
    Factory class to conveniently acces the different
    fitting algorithms available in atomicrex.
    """

    @staticmethod
    def ar_lbfgs(conv_threshold=1e-10, max_iter=50, gradient_epsilon=None):
        return AtomicrexAlgorithm(
            name="BFGS",
            conv_threshold=conv_threshold,
            max_iter=max_iter,
            gradient_epsilon=gradient_epsilon,
        )

    @staticmethod
    def ar_spa(spa_iterations=20, seed=42):
        return SpaMinimizer(spa_iterations=spa_iterations, seed=seed)

    @staticmethod
    def ld_lbfgs(
        stopval=1e-10,
        max_iter=50,
        maxtime=None,
        ftol_rel=None,
        ftol_abs=None,
        xtol_rel=None,
        seed=None,
    ):
        return NloptAlgorithm(
            name="LD_LBFGS",
            stopval=stopval,
            max_iter=max_iter,
            maxtime=maxtime,
            ftol_rel=ftol_rel,
            ftol_abs=ftol_abs,
            xtol_rel=xtol_rel,
            seed=seed,
        )

    @staticmethod
    def ld_mma(
        stopval=1e-10,
        max_iter=50,
        maxtime=None,
        ftol_rel=None,
        ftol_abs=None,
        xtol_rel=None,
        seed=None,
    ):
        return NloptAlgorithm(
            name="LD_MMA",
            stopval=stopval,
            max_iter=max_iter,
            maxtime=maxtime,
            ftol_rel=ftol_rel,
            ftol_abs=ftol_abs,
            xtol_rel=xtol_rel,
            seed=seed,
        )

    @staticmethod
    def ld_ccsaq(
        stopval=1e-10,
        max_iter=50,
        maxtime=None,
        ftol_rel=None,
        ftol_abs=None,
        xtol_rel=None,
        seed=None,
    ):
        return NloptAlgorithm(
            name="LD_CCSAQ",
            stopval=stopval,
            max_iter=max_iter,
            maxtime=maxtime,
            ftol_rel=ftol_rel,
            ftol_abs=ftol_abs,
            xtol_rel=xtol_rel,
            seed=seed,
        )

    @staticmethod
    def ld_slsqp(
        stopval=1e-10,
        max_iter=50,
        maxtime=None,
        ftol_rel=None,
        ftol_abs=None,
        xtol_rel=None,
        seed=None,
    ):
        return NloptAlgorithm(
            name="LD_SLSQP",
            stopval=stopval,
            max_iter=max_iter,
            maxtime=maxtime,
            ftol_rel=ftol_rel,
            ftol_abs=ftol_abs,
            xtol_rel=xtol_rel,
            seed=seed,
        )

    @staticmethod
    def ld_var1(
        stopval=1e-10,
        max_iter=50,
        maxtime=None,
        ftol_rel=None,
        ftol_abs=None,
        xtol_rel=None,
        seed=None,
    ):
        return NloptAlgorithm(
            name="LD_VAR1",
            stopval=stopval,
            max_iter=max_iter,
            maxtime=maxtime,
            ftol_rel=ftol_rel,
            ftol_abs=ftol_abs,
            xtol_rel=xtol_rel,
            seed=seed,
        )

    @staticmethod
    def ld_var2(
        stopval=1e-10,
        max_iter=50,
        maxtime=None,
        ftol_rel=None,
        ftol_abs=None,
        xtol_rel=None,
        seed=None,
    ):
        return NloptAlgorithm(
            name="LD_VAR2",
            stopval=stopval,
            max_iter=max_iter,
            maxtime=maxtime,
            ftol_rel=ftol_rel,
            ftol_abs=ftol_abs,
            xtol_rel=xtol_rel,
            seed=seed,
        )

    @staticmethod
    def ln_cobyla(
        stopval=1e-10,
        max_iter=50,
        maxtime=None,
        ftol_rel=None,
        ftol_abs=None,
        xtol_rel=None,
        seed=None,
    ):
        return NloptAlgorithm(
            name="LN_COBYLA",
            stopval=stopval,
            max_iter=max_iter,
            maxtime=maxtime,
            ftol_rel=ftol_rel,
            ftol_abs=ftol_abs,
            xtol_rel=xtol_rel,
            seed=seed,
        )

    @staticmethod
    def ln_bobyqa(
        stopval=1e-10,
        max_iter=50,
        maxtime=None,
        ftol_rel=None,
        ftol_abs=None,
        xtol_rel=None,
        seed=None,
    ):
        return NloptAlgorithm(
            name="LN_BOBYQA",
            stopval=stopval,
            max_iter=max_iter,
            maxtime=maxtime,
            ftol_rel=ftol_rel,
            ftol_abs=ftol_abs,
            xtol_rel=xtol_rel,
            seed=seed,
        )

    @staticmethod
    def ln_newuoa(
        stopval=1e-10,
        max_iter=50,
        maxtime=None,
        ftol_rel=None,
        ftol_abs=None,
        xtol_rel=None,
        seed=None,
    ):
        return NloptAlgorithm(
            name="LN_NEWUOA",
            stopval=stopval,
            max_iter=max_iter,
            maxtime=maxtime,
            ftol_rel=ftol_rel,
            ftol_abs=ftol_abs,
            xtol_rel=xtol_rel,
            seed=seed,
        )

    @staticmethod
    def ln_newuoa_bound(
        stopval=1e-10,
        max_iter=50,
        maxtime=None,
        ftol_rel=None,
        ftol_abs=None,
        xtol_rel=None,
        seed=None,
    ):
        return NloptAlgorithm(
            name="LN_NEWUOA_BOUND",
            stopval=stopval,
            max_iter=max_iter,
            maxtime=maxtime,
            ftol_rel=ftol_rel,
            ftol_abs=ftol_abs,
            xtol_rel=xtol_rel,
            seed=seed,
        )

    @staticmethod
    def ln_praxis(
        stopval=1e-10,
        max_iter=50,
        maxtime=None,
        ftol_rel=None,
        ftol_abs=None,
        xtol_rel=None,
        seed=None,
    ):
        return NloptAlgorithm(
            name="LN_PRAXIS",
            stopval=stopval,
            max_iter=max_iter,
            maxtime=maxtime,
            ftol_rel=ftol_rel,
            ftol_abs=ftol_abs,
            xtol_rel=xtol_rel,
            seed=seed,
        )

    @staticmethod
    def ln_neldermead(
        stopval=1e-10,
        max_iter=50,
        maxtime=None,
        ftol_rel=None,
        ftol_abs=None,
        xtol_rel=None,
        seed=None,
    ):
        return NloptAlgorithm(
            name="LN_NELDERMEAD",
            stopval=stopval,
            max_iter=max_iter,
            maxtime=maxtime,
            ftol_rel=ftol_rel,
            ftol_abs=ftol_abs,
            xtol_rel=xtol_rel,
            seed=seed,
        )

    @staticmethod
    def ln_sbplx(
        stopval=1e-10,
        max_iter=50,
        maxtime=None,
        ftol_rel=None,
        ftol_abs=None,
        xtol_rel=None,
        seed=None,
    ):
        return NloptAlgorithm(
            name="LN_SBPLX",
            stopval=stopval,
            max_iter=max_iter,
            maxtime=maxtime,
            ftol_rel=ftol_rel,
            ftol_abs=ftol_abs,
            xtol_rel=xtol_rel,
            seed=seed,
        )

    @staticmethod
    def gn_crs2_lm(
        stopval=1e-10,
        max_iter=50,
        maxtime=None,
        ftol_rel=None,
        ftol_abs=None,
        xtol_rel=None,
        seed=None,
    ):
        return NloptAlgorithm(
            name="GN_CRS2_LM",
            stopval=stopval,
            max_iter=max_iter,
            maxtime=maxtime,
            ftol_rel=ftol_rel,
            ftol_abs=ftol_abs,
            xtol_rel=xtol_rel,
            seed=seed,
        )

    @staticmethod
    def gn_esch(
        stopval=1e-10,
        max_iter=50,
        maxtime=None,
        ftol_rel=None,
        ftol_abs=None,
        xtol_rel=None,
        seed=None,
    ):
        return NloptAlgorithm(
            name="GN_ESCH",
            stopval=stopval,
            max_iter=max_iter,
            maxtime=maxtime,
            ftol_rel=ftol_rel,
            ftol_abs=ftol_abs,
            xtol_rel=xtol_rel,
            seed=seed,
        )

    @staticmethod
    def gn_direct(
        stopval=1e-10,
        max_iter=50,
        maxtime=None,
        ftol_rel=None,
        ftol_abs=None,
        xtol_rel=None,
        seed=None,
    ):
        return NloptAlgorithm(
            name="GN_DIRECT",
            stopval=stopval,
            max_iter=max_iter,
            maxtime=maxtime,
            ftol_rel=ftol_rel,
            ftol_abs=ftol_abs,
            xtol_rel=xtol_rel,
            seed=seed,
        )

    @staticmethod
    def gn_direct_l(
        stopval=1e-10,
        max_iter=50,
        maxtime=None,
        ftol_rel=None,
        ftol_abs=None,
        xtol_rel=None,
        seed=None,
    ):
        return NloptAlgorithm(
            name="GN_DIRECT_L",
            stopval=stopval,
            max_iter=max_iter,
            maxtime=maxtime,
            ftol_rel=ftol_rel,
            ftol_abs=ftol_abs,
            xtol_rel=xtol_rel,
            seed=seed,
        )

    @staticmethod
    def gn_isres(
        stopval=1e-10,
        max_iter=50,
        maxtime=None,
        ftol_rel=None,
        ftol_abs=None,
        xtol_rel=None,
        seed=None,
    ):
        return NloptAlgorithm(
            name="GN_ISRES",
            stopval=stopval,
            max_iter=max_iter,
            maxtime=maxtime,
            ftol_rel=ftol_rel,
            ftol_abs=ftol_abs,
            xtol_rel=xtol_rel,
            seed=seed,
        )

    @staticmethod
    def g_mlsl(
        stopval=1e-10,
        max_iter=50,
        maxtime=None,
        ftol_rel=None,
        ftol_abs=None,
        xtol_rel=None,
        seed=None,
    ):
        return NloptGlobalLocal(
            name="G_MLSL",
            stopval=stopval,
            max_iter=max_iter,
            maxtime=maxtime,
            ftol_rel=ftol_rel,
            ftol_abs=ftol_abs,
            xtol_rel=xtol_rel,
            seed=seed,
        )

    @staticmethod
    def g_mlsl_lds(
        stopval=1e-10,
        max_iter=50,
        maxtime=None,
        ftol_rel=None,
        ftol_abs=None,
        xtol_rel=None,
        seed=None,
    ):
        return NloptGlobalLocal(
            name="G_MLSL_LDS",
            stopval=stopval,
            max_iter=max_iter,
            maxtime=maxtime,
            ftol_rel=ftol_rel,
            ftol_abs=ftol_abs,
            xtol_rel=xtol_rel,
            seed=seed,
        )

    @staticmethod
    def gd_stogo(
        stopval=1e-10,
        max_iter=50,
        maxtime=None,
        ftol_rel=None,
        ftol_abs=None,
        xtol_rel=None,
        seed=None,
    ):
        return NloptAlgorithm(
            name="GD_STOGO",
            stopval=stopval,
            max_iter=max_iter,
            maxtime=maxtime,
            ftol_rel=ftol_rel,
            ftol_abs=ftol_abs,
            xtol_rel=xtol_rel,
            seed=seed,
        )

    @staticmethod
    def gd_stogo_rand(
        stopval=1e-10,
        max_iter=50,
        maxtime=None,
        ftol_rel=None,
        ftol_abs=None,
        xtol_rel=None,
        seed=None,
    ):
        return NloptAlgorithm(
            name="GD_STOGO_RAND",
            stopval=stopval,
            max_iter=max_iter,
            maxtime=maxtime,
            ftol_rel=ftol_rel,
            ftol_abs=ftol_abs,
            xtol_rel=xtol_rel,
            seed=seed,
        )

    @staticmethod
    def scipy_algorithm():
        return ScipyAlgorithm()


class AtomicrexAlgorithm(DataContainer):
    """
    Class to inherit from. If more algorithms will be implemented in atomicrex
    at some point they can be implemented similar to the AR_LBFGS class.
    """

    def __init__(
        self,
        conv_threshold=None,
        max_iter=None,
        gradient_epsilon=None,
        name=None,
        *args,
        **kwargs,
    ):
        super().__init__(table_name="fitting_algorithm", *args, **kwargs)
        self.conv_threshold = conv_threshold
        self.max_iter = max_iter
        self.gradient_epsilon = gradient_epsilon
        self.name = name

    def _to_xml_element(self):
        """Internal function.
        Converts the algorithm to an xml element
        """
        algo = ET.Element(self.name)
        algo.set("conv-threshold", f"{self.conv_threshold}")
        algo.set("max-iter", f"{self.max_iter}")
        if self.gradient_epsilon is not None:
            algo.set("gradient-epsilon", f"{self.gradient_epsilon}")
        return algo


class SpaMinimizer(DataContainer):
    """
    Global optimizer implemented in atomicrex.
    Should be used in combination with a local minimizer.
    See the atomicrex documentation for details.
    """

    def __init__(self, max_iter=None, seed=None, *args, **kwargs):
        super().__init__(table_name="fitting_algorithm", *args, **kwargs)
        self.max_iter = max_iter
        self.seed = seed
        self.local_minimizer = None
    
    @property
    def name(self):
        return "spa"

    def _to_xml_element(self):
        """Internal function.
        Converts the algorithm to a xml element
        and returns it
        """
        spa = ET.Element("spa")
        spa.set("max-iter", f"{self.max_iter}")
        spa.set("seed", f"{self.seed}")
        if self.local_minimizer is not None:
            spa.append(self.local_minimizer._to_xml_element())
        else:
            raise ValueError("Set a local minimizer for Spa")
        return spa


class NloptAlgorithm(DataContainer):
    """
    Nlopt algorithms should inherit from this class.
    """

    def __init__(
        self,
        name=None,
        seed=None,
        stopval=None,
        max_iter=None,
        maxtime=None,
        ftol_rel=None,
        ftol_abs=None,
        xtol_rel=None,
        *args,
        **kwargs,
    ):
        super().__init__(table_name="fitting_algorithm", *args, **kwargs)
        self.stopval = stopval
        self.max_iter = max_iter
        self.maxtime = maxtime
        self.ftol_rel = ftol_rel
        self.ftol_abs = ftol_abs
        self.xtol_rel = xtol_rel
        self.seed = seed
        self.name = name

    def _to_xml_element(self):
        """Internal Function.
        Converts the algorithm to a xml element
        and returns it
        """
        nlopt = ET.Element("nlopt")
        nlopt.set("algorithm", self.name)
        if self.stopval is not None:
            nlopt.set("stopval", f"{self.stopval}")
        if self.max_iter is not None:
            nlopt.set("maxeval", f"{self.max_iter}")
        if self.maxtime is not None:
            nlopt.set("maxtime", f"{self.maxtime}")
        if self.ftol_rel is not None:
            nlopt.set("ftol_rel", f"{self.ftol_rel}")
        if self.ftol_abs is not None:
            nlopt.set("ftol_abs", f"{self.ftol_abs}")
        if self.xtol_rel is not None:
            nlopt.set("xtol_rel", f"{self.xtol_rel}")
        if self.seed is not None:
            nlopt.set("seed", f"{self.seed}")
        return nlopt


class NloptGlobalLocal(NloptAlgorithm):
    """
    Nlopt global optimizers that additionally need a local minimizer similar to the spa algorithm implemented in atomicrex
    """

    def __init__(
        self,
        stopval=None,
        max_iter=None,
        maxtime=None,
        ftol_rel=None,
        ftol_abs=None,
        xtol_rel=None,
        name=None,
        seed=None,
        *args,
        **kwargs,
    ):
        super().__init__(
            stopval=stopval,
            max_iter=max_iter,
            maxtime=maxtime,
            ftol_rel=ftol_rel,
            ftol_abs=ftol_abs,
            xtol_rel=xtol_rel,
            name=name,
            seed=seed,
            *args,
            **kwargs,
        )
        self.local_minimizer = None

    def _to_xml_element(self):
        """Internal Function.
        Converts the algorithm to a xml element
        and returns it
        """
        if self.local_minimizer is None:
            raise ValueError("This global minimzer needs an additional local minimzer")
        nlopt = super()._to_xml_element()
        local = self.local_minimizer._to_xml_element()
        nlopt.append(local)
        return nlopt


class ScipyAlgorithm:
    __version__ = "0.0.1"
    __hdf_version__ = "0.0.1"

    def __init__(self):
        self.global_minimizer = None
        self.local_minimizer_kwargs = {
            "method": "L-BFGS-B",
            "jac": None,
            "hess": None,
            "bounds": None,
            "constraints": (),
            "tol": None,
            "options": None,
        }
        self.global_minimizer_kwargs = {}

    def to_hdf(self, hdf, group_name):
        with hdf.open(group_name) as h:
            self._type_to_hdf(h)
            h["global_minimizer"] = self.global_minimizer
        """
            with h.open("local_minimizer_kwargs") as loc_hdf:
                for k, v in self.local_minimizer_kwargs.items():
                    try:
                        loc_hdf[k] = v
                    except TypeError:
                        loc_hdf[k] = v.__name__
            with h.open("global_minimizer_kwargs") as glob_hdf:
                for k, v in self.global_minimizer_kwargs.items():
                    if isinstance(v, dict):
                        with glob_hdf.open(k) as v_hdf:
                            for k, v in self.v.items():
                                try:
                                    v_hdf[k] = v
                                except TypeError:
                                    v_hdf[k] = v.__name__
                    else:           
                        try:
                            glob_hdf[k] = v
                        except TypeError:
                            glob_hdf[k] = v.__name__
        """

    def from_hdf(self, hdf, group_name):
        with hdf.open(group_name) as h:
            #self._type_from_hdf(h)
            self.global_minimizer = h["global_minimizer"]

    def _type_to_hdf(self, hdf):
        """
        Internal helper function to save type and version in hdf root

        Args:
            hdf (ProjectHDFio): HDF5 group object
        """
        hdf["NAME"] = self.__class__.__name__
        hdf["TYPE"] = str(type(self))
        hdf["VERSION"] = self.__version__
        hdf["HDF_VERSION"] = self.__hdf_version__
